"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
import pytest
from onnx import numpy_helper
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.passes.canonicalization.reset_reshape_batch import trace_batch_dimension


def _make_graph():
    # a.T + reshape(b) + c.squeeze(1) + d.unsqueeze(-1) + gather(e)
    a = make_node("Transpose", ["a"], ["at"], perm=[0, 2, 1], name="at")
    b = make_node("Reshape", ["b", "bs"], ["br"], name="br")
    c = make_node("Squeeze", ["c", "ca"], ["cs"], name="cs")
    d = make_node("Unsqueeze", ["d", "da"], ["du"], name="du")
    e = make_node("Gather", ["e", "ei"], ["eg"], axis=3, name="eg")
    ca = make_node("Constant", [], ["ca"], value_int=1)
    add = make_node("Sum", ["at", "br", "cs", "du", "eg"], ["y"], name="add")
    graph = make_graph(
        [a, b, ca, c, d, e, add],
        "test",
        [
            make_tensor_value_info("a", onnx.TensorProto.FLOAT, [1, 128, 32]),
            make_tensor_value_info("b", onnx.TensorProto.FLOAT, [1, 32 * 128]),
            make_tensor_value_info("c", onnx.TensorProto.FLOAT, [1, 1, 1, 128]),
            make_tensor_value_info("d", onnx.TensorProto.FLOAT, [1, 1]),
            make_tensor_value_info("e", onnx.TensorProto.FLOAT, [1, 32, 1, 2]),
        ],
        [
            make_tensor_value_info("y", onnx.TensorProto.FLOAT, [1, 32, 128]),
        ],
        [
            numpy_helper.from_array(np.array([1, 32, 128], np.int64), "bs"),
            # numpy_helper.from_array(np.array(1, np.int64), "ca"),
            numpy_helper.from_array(np.array([-1], np.int64), "da"),
            numpy_helper.from_array(np.array(0, np.int64), "ei"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )

    onnx.checker.check_model(model, True)
    return OnnxGraph(model)


def _make_graph2():
    a = make_node("Transpose", ["a"], ["at"], perm=[1, 2, 0], name="at")
    b = make_node("Reshape", ["at", "bs"], ["br"], name="br")
    graph = make_graph(
        [a, b],
        "test",
        [make_tensor_value_info("a", onnx.TensorProto.FLOAT, [1, 32, 128])],
        [make_tensor_value_info("br", onnx.TensorProto.FLOAT, [32 * 128, 1])],
        [numpy_helper.from_array(np.array([32 * 128, 1], np.int64), "bs")],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return OnnxGraph(model)


def _make_graph3():
    a = make_node("Transpose", ["a"], ["at"], perm=[1, 2, 0], name="at")
    b = make_node("Reshape", ["at", "bs"], ["br"], name="br")
    graph = make_graph(
        [a, b],
        "test",
        [make_tensor_value_info("a", onnx.TensorProto.FLOAT, [2, 128, 32])],
        [make_tensor_value_info("br", onnx.TensorProto.FLOAT, [128, 64])],
        [numpy_helper.from_array(np.array([128, 64], np.int64), "bs")],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return OnnxGraph(model)


def _make_bad_graph():
    a = make_node("Transpose", ["a"], ["at"], perm=[1, 2, 0], name="at")
    b = make_node("Reshape", ["at", "bs"], ["br"], name="br")
    graph = make_graph(
        [a, b],
        "test",
        [make_tensor_value_info("a", onnx.TensorProto.FLOAT, [2, 32, 128])],
        [make_tensor_value_info("br", onnx.TensorProto.FLOAT, [64, 128])],
        [numpy_helper.from_array(np.array([64, 128], np.int64), "bs")],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return OnnxGraph(model)


def _make_bad_graph2():
    a = make_node("Transpose", ["a"], ["at"], perm=[1, 2, 0], name="at")
    b = make_node("Reshape", ["at", "bs"], ["br"], name="br")
    graph = make_graph(
        [a, b],
        "test",
        [make_tensor_value_info("a", onnx.TensorProto.FLOAT, [2, 32, 128])],
        [make_tensor_value_info("br", onnx.TensorProto.FLOAT, [32, 2, 128])],
        [numpy_helper.from_array(np.array([32, 2, 128], np.int64), "bs")],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return OnnxGraph(model)


def test_trace_batch_dimension():
    graph = _make_graph()
    tracing = trace_batch_dimension(graph)
    assert tracing["at"] == {0}
    assert tracing["br"] == {0}
    assert tracing["cs"] == {0}
    assert tracing["du"] == {0}
    assert tracing["eg"] == {0}
    assert tracing["add"] == {0}


def test_reset_reshape_batch():
    graph = _make_graph()
    pm = PassManager(["reset_reshape_batch"])
    graph = pm.optimize(graph, strict=True)

    reshape = graph.onnx_predecessors(graph.nodes["br"]["pb"])[-1].attribute[0].t
    assert -1 == numpy_helper.to_array(reshape)[0]


def test_reset_reshape_batch_to_last_dim():
    graph = _make_graph2()
    pm = PassManager(["reset_reshape_batch"])
    graph = pm.optimize(graph, strict=True)

    reshape = graph.onnx_predecessors(graph.nodes["br"]["pb"])[-1].attribute[0].t
    assert -1 == numpy_helper.to_array(reshape)[-1]


def test_reset_reshape_batch_merged_dim():
    graph = _make_graph3()
    pm = PassManager(["reset_reshape_batch"])
    graph = pm.optimize(graph, strict=True)

    reshape = graph.onnx_predecessors(graph.nodes["br"]["pb"])[-1].attribute[0].t
    assert -1 == numpy_helper.to_array(reshape)[-1]


def test_reset_reshape_batch_bad_graph():
    pm = PassManager(["reset_reshape_batch"])

    graph = _make_bad_graph()
    with pytest.raises((ValueError, NotImplementedError)):
        pm.optimize(graph, strict=True)

    graph = _make_bad_graph2()
    with pytest.raises((ValueError, NotImplementedError)):
        pm.optimize(graph, strict=True)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Mapping, MutableMapping, TypeAlias, Union

from onnx import NodeProto
from onnx.helper import make_function, make_node, make_operatorsetid

from ...graph import OnnxGraph
from .. import PASSES

DOMAIN = "org.onnxifier"
HierType: TypeAlias = Mapping[str, Union[str, "HierType"]]
MutableHierType: TypeAlias = MutableMapping[str, Union[str, "MutableHierType"]]


def _group_operators_recursively(
    graph: OnnxGraph, hier: HierType, group_name: str
) -> NodeProto:
    nodes = []
    for name, hierarchy in hier.items():
        if isinstance(hierarchy, Mapping):
            # subgraph node
            node = _group_operators_recursively(graph, hierarchy, name)
        else:
            node = graph.nodes[hierarchy]["pb"]
        # fuse constants
        for pred in graph.onnx_predecessors(node):
            if pred.op_type == "Constant":
                nodes.append(pred)
        nodes.append(node)
    h = graph.onnx_subgraph(nodes)
    h.name = group_name
    func = make_function(
        DOMAIN,
        group_name,
        inputs=[i.name for i in h.input],
        outputs=[i.name for i in h.output],
        nodes=nodes,
        opset_imports=[
            make_operatorsetid(DOMAIN, graph.opset_version),
            make_operatorsetid("", graph.opset_version),
        ],
    )
    graph.onnx_add_function(func)
    hnode = make_node(
        op_type=group_name,
        inputs=[i.name for i in h.input],
        outputs=[i.name for i in h.output],
        name=group_name,
        domain=DOMAIN,
    )
    graph.add_onnx_node(hnode)
    for n in nodes:
        graph.remove_onnx_node(n)
    return hnode


@PASSES.register("group", deps=["initializer_to_constant"])
def group_operators(graph: OnnxGraph, depth: int = 1, sep: str = "/") -> OnnxGraph:
    """Group operators into nested subgraphs (functions) based on their name.

    Note:

        Prefix is obtained by splitting name by `sep` and taking first `depth` parts.

    Example:

        model/backbone/layer1/conv1
            -> model -> /backbone/layer1/conv1 (depth=1)
            -> model -> /backbone -> /layer1/conv1 (depth=2)
            -> model -> /backbone -> /layer1 -> /conv1 (depth>=3)
    """

    assert depth >= 1 and len(sep) == 1
    hier: MutableHierType = {}
    for node in graph:
        node_pb = graph.nodes[node]["pb"]
        if not node_pb.HasField("name") or node_pb.op_type == "Constant":
            continue
        name = node_pb.name.strip(sep)  # remove leading and trailing slashes
        if sep not in name:
            continue
        # remove leading and trailing slashes, then split by sep
        *parents, leaf = name.split(sep, depth)

        curr = hier
        for parent in parents:
            # Detect naming collisions where a path segment was previously used
            # as a leaf node (string) but now needs to become a subtree.
            if isinstance(curr, str):
                raise ValueError(
                    f"Naming collision in group_operators: path segment '{parent}' "
                    f"was previously used as a leaf node when processing '{name}'."
                )
            if parent not in curr:
                curr[parent] = {}
            elif isinstance(curr[parent], str):
                # Existing leaf at this path cannot be converted into a subtree
                # without changing semantics; report a clear error instead.
                full_path = sep.join((*parents[: parents.index(parent) + 1],))
                raise ValueError(
                    "Naming collision in group_operators: path segment "
                    f"'{full_path}' is used both as an operator name and as a "
                    f"prefix for other operators (while processing '{name}')."
                )
            curr = curr[parent]
        assert isinstance(curr, dict)
        curr[leaf] = node

    for name, hierarchy in hier.items():
        if isinstance(hierarchy, Mapping):
            _group_operators_recursively(graph, hierarchy, name)

    return graph



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List

import numpy as np
from onnx.helper import make_tensor_type_proto, make_value_info

from ...graph import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(
    name="expand_conv_channel", deps=["initializer_to_constant", "infer_shape"]
)
class ExpandConvChannelRewriter(Rewriter):
    """Expand Conv's channel to 4"""

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("Conv"))

    def rewrite(self, graph: OnnxGraph, nodes: List):
        conv_node = nodes[0]
        n, ic, h, w = graph.tensor_shape(conv_node.input[0])
        if ic != 3:
            return
        old_inputs = list(filter(lambda x: x.name == conv_node.input[0], graph.input))
        if len(old_inputs) != 1:
            return
        old_input = old_inputs[0]
        # make sure that only one node connects with the old_input
        # TODO: support that multiple ops use old_input
        nodes = list(
            filter(lambda x: old_input.name in x["pb"].input, graph.nodes.values())
        )
        if len(nodes) != 1:
            return
        # make sure that only one node connects with the old_input
        # TODO: support that multiple ops use old_input
        nodes = list(
            filter(lambda x: old_input.name in x["pb"].input, graph.nodes.values())
        )
        if len(nodes) != 1:
            return

        # expand weight
        weight_node = self.get_input_node_or_die(conv_node, 1)
        weight_value = self.get_value_or_die(weight_node)
        oc, _, kh, kw = weight_value.shape
        expand_weight_value = np.zeros([oc, 4, kh, kw], dtype=weight_value.dtype)
        expand_weight_value[:, :3, :, :] = weight_value

        expand_weight_node = make_constant(
            weight_node.name + "/expand", expand_weight_value
        )
        conv_node.input[0] += "_expand"
        conv_node.input[1] = expand_weight_node.output[0]

        # replace graph input
        graph.input.append(
            make_value_info(
                conv_node.input[0],
                make_tensor_type_proto(1, [n, 4, h, w]),
            )
        )
        graph.input.remove(old_input)
        graph.inputs.pop(old_input.name)
        self -= weight_node
        self += expand_weight_node



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info

from onnxifier import OnnxGraph, PassManager


def test_constant_to_constantofshape_fp32():
    """Test conversion of fp32 constant to ConstantOfShape with threshold."""
    # Create a constant node with enough elements to exceed default threshold (16)
    constant_value = np.full((4, 5), 1.5, dtype=np.float32)  # 20 elements > 16
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.FLOAT, [4, 5])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass
    onnx_graph = OnnxGraph(model)
    pm = PassManager(["constant_to_constantofshape"])
    optimized_graph = pm.optimize(onnx_graph, strict=True)

    # Verify the result
    nodes = [optimized_graph.nodes[node]["pb"] for node in optimized_graph.nodes]

    # Should have two nodes: one shape constant and one ConstantOfShape
    assert len(nodes) == 2

    shape_nodes = [n for n in nodes if n.op_type == "Constant"]
    constantofshape_nodes = [n for n in nodes if n.op_type == "ConstantOfShape"]

    assert len(shape_nodes) == 1
    assert len(constantofshape_nodes) == 1

    # Verify the ConstantOfShape node has correct value
    cos_node = constantofshape_nodes[0]
    value_attr = next(attr for attr in cos_node.attribute if attr.name == "value")
    value_array = onnx.numpy_helper.to_array(value_attr.t)
    assert value_array.dtype == np.float32
    assert value_array[0] == 1.5


def test_constant_to_constantofshape_fp16():
    """Test conversion of fp16 constant to ConstantOfShape with threshold."""
    # Create a constant node with enough elements to exceed default threshold (16)
    constant_value = np.full((3, 6), 2.5, dtype=np.float16)  # 18 elements > 16
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.FLOAT16, [3, 6])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass
    onnx_graph = OnnxGraph(model)
    pm = PassManager(["constant_to_constantofshape"])
    optimized_graph = pm.optimize(onnx_graph, strict=True)

    # Verify the result
    nodes = [optimized_graph.nodes[node]["pb"] for node in optimized_graph.nodes]

    # Should have two nodes: one shape constant and one ConstantOfShape
    assert len(nodes) == 2

    constantofshape_nodes = [n for n in nodes if n.op_type == "ConstantOfShape"]
    assert len(constantofshape_nodes) == 1

    # Verify the ConstantOfShape node has correct value
    cos_node = constantofshape_nodes[0]
    value_attr = next(attr for attr in cos_node.attribute if attr.name == "value")
    value_array = onnx.numpy_helper.to_array(value_attr.t)
    assert value_array.dtype == np.float16
    assert value_array[0] == np.float16(2.5)


def test_constant_to_constantofshape_int32():
    """Test conversion of int32 constant to ConstantOfShape with threshold."""
    # Create a constant node with enough elements to exceed default threshold (16)
    constant_value = np.full((4, 5), 42, dtype=np.int32)  # 20 elements > 16
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.INT32, [4, 5])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass
    onnx_graph = OnnxGraph(model)
    pm = PassManager(["constant_to_constantofshape"])
    optimized_graph = pm.optimize(onnx_graph, strict=True)

    # Verify the result
    nodes = [optimized_graph.nodes[node]["pb"] for node in optimized_graph.nodes]

    # Should have two nodes: one shape constant and one ConstantOfShape
    assert len(nodes) == 2

    constantofshape_nodes = [n for n in nodes if n.op_type == "ConstantOfShape"]
    assert len(constantofshape_nodes) == 1

    # Verify the ConstantOfShape node has correct value
    cos_node = constantofshape_nodes[0]
    value_attr = next(attr for attr in cos_node.attribute if attr.name == "value")
    value_array = onnx.numpy_helper.to_array(value_attr.t)
    assert value_array.dtype == np.int32
    assert value_array[0] == 42


def test_constant_to_constantofshape_int64():
    """Test conversion of int64 constant to ConstantOfShape with threshold."""
    # Create a constant node with enough elements to exceed default threshold (16)
    constant_value = np.full((3, 6), 123456789, dtype=np.int64)  # 18 elements > 16
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.INT64, [3, 6])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass
    onnx_graph = OnnxGraph(model)
    pm = PassManager(["constant_to_constantofshape"])
    optimized_graph = pm.optimize(onnx_graph, strict=True)

    # Verify the result
    nodes = [optimized_graph.nodes[node]["pb"] for node in optimized_graph.nodes]

    # Should have two nodes: one shape constant and one ConstantOfShape
    assert len(nodes) == 2

    constantofshape_nodes = [n for n in nodes if n.op_type == "ConstantOfShape"]
    assert len(constantofshape_nodes) == 1

    # Verify the ConstantOfShape node has correct value
    cos_node = constantofshape_nodes[0]
    value_attr = next(attr for attr in cos_node.attribute if attr.name == "value")
    value_array = onnx.numpy_helper.to_array(value_attr.t)
    assert value_array.dtype == np.int64
    assert value_array[0] == 123456789


def test_constant_to_constantofshape_skip_under_threshold():
    """Test that constants below threshold are not converted."""
    # Create a constant node with elements below threshold
    # 4 elements < 16
    constant_value = np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float32)
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.FLOAT, [2, 2])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass
    onnx_graph = OnnxGraph(model)
    pm = PassManager(["constant_to_constantofshape"])
    optimized_graph = pm.optimize(onnx_graph, strict=True)

    # Verify the result - should remain unchanged
    nodes = [optimized_graph.nodes[node]["pb"] for node in optimized_graph.nodes]
    assert len(nodes) == 1
    assert nodes[0].op_type == "Constant"


def test_constant_to_constantofshape_scalar():
    """Test that scalar constants below threshold are not converted."""
    # Create a scalar constant node (1 element < 16)
    constant_value = np.array(3.14, dtype=np.float32)
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.FLOAT, [])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass
    onnx_graph = OnnxGraph(model)
    pm = PassManager(["constant_to_constantofshape"])
    optimized_graph = pm.optimize(onnx_graph, strict=True)

    # Verify the result - should remain unchanged (below threshold)
    nodes = [optimized_graph.nodes[node]["pb"] for node in optimized_graph.nodes]
    assert len(nodes) == 1
    assert nodes[0].op_type == "Constant"


def test_constant_to_constantofshape_with_custom_threshold():
    """Test conversion with custom threshold."""
    # Create a constant node with 10 elements
    constant_value = np.full((2, 5), 1.5, dtype=np.float32)  # 10 elements
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.FLOAT, [2, 5])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass with threshold=5 (should convert)
    onnx_graph = OnnxGraph(model)
    from onnxifier.passes.swap.constant_to_constantofshape import (
        ConstantToConstantOfShapeRewriter,
    )

    # Create rewriter with custom threshold
    rewriter = ConstantToConstantOfShapeRewriter(threshold=5)
    onnx_graph = rewriter.match_and_rewrite(onnx_graph)

    # Verify the result - should be converted
    nodes = [onnx_graph.nodes[node]["pb"] for node in onnx_graph.nodes]
    assert len(nodes) == 2

    constantofshape_nodes = [n for n in nodes if n.op_type == "ConstantOfShape"]
    assert len(constantofshape_nodes) == 1

    # Verify the ConstantOfShape node has correct value (first element)
    cos_node = constantofshape_nodes[0]
    value_attr = next(attr for attr in cos_node.attribute if attr.name == "value")
    value_array = onnx.numpy_helper.to_array(value_attr.t)
    assert value_array.dtype == np.float32
    assert value_array[0] == 1.5


def test_constant_to_constantofshape_takes_first_element():
    """Test that conversion takes the first element regardless of uniformity."""
    # Create a constant node with non-uniform values above threshold
    constant_value = np.array(
        [
            [1.0, 2.0, 3.0, 4.0, 5.0],
            [6.0, 7.0, 8.0, 9.0, 10.0],
            [11.0, 12.0, 13.0, 14.0, 15.0],
            [16.0, 17.0, 18.0, 19.0, 20.0],
        ],
        dtype=np.float32,
    )  # 20 elements > 16, non-uniform
    constant_node = make_node(
        "Constant",
        inputs=[],
        outputs=["const_output"],
        value=onnx.numpy_helper.from_array(constant_value, "const_value"),
    )

    graph = make_graph(
        [constant_node],
        "test_graph",
        [],
        [make_tensor_value_info("const_output", onnx.TensorProto.FLOAT, [4, 5])],
    )
    model = make_model(graph)
    onnx.checker.check_model(model, True)

    # Apply the pass
    onnx_graph = OnnxGraph(model)
    pm = PassManager(["constant_to_constantofshape"])
    optimized_graph = pm.optimize(onnx_graph, strict=True)

    # Verify the result - should be converted
    nodes = [optimized_graph.nodes[node]["pb"] for node in optimized_graph.nodes]
    assert len(nodes) == 2

    constantofshape_nodes = [n for n in nodes if n.op_type == "ConstantOfShape"]
    assert len(constantofshape_nodes) == 1

    # Verify the ConstantOfShape node has the first element value (1.0)
    cos_node = constantofshape_nodes[0]
    value_attr = next(attr for attr in cos_node.attribute if attr.name == "value")
    value_array = onnx.numpy_helper.to_array(value_attr.t)
    assert value_array.dtype == np.float32
    assert value_array[0] == 1.0  # First element of the original array



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from openvino import Model
from openvino.opset1 import constant, parameter
from openvino.opset4 import scatter_nd_update

from . import convert_xml


def test_scatter_nd_update_opset4():
    data = parameter([8])
    indices = constant(np.array([[4], [3], [1], [7]], np.int64))
    updates = constant(np.array([9, 10, 11, 12], np.float32))
    r = scatter_nd_update(data, indices, updates)
    m = Model([r], [data])
    convert_xml(m)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from .. import OnnxGraph, SingleNodePattern
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="z__trivial")  # make sure to be called in the last
class Trivial(BaseNodeConversion):
    """Trivial conversion for OpenVINO nodes."""

    def __init__(self):
        super().__init__(SingleNodePattern().with_attr("version"))

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        # replace ai.intel.openvino domain to ai.onnx (empty)
        if ori_node.domain != "ai.intel.openvino":
            raise RuntimeError(
                f"node {ori_node.name} got an unexpected domain: {ori_node.domain}"
            )
        return make_node(
            ori_node.op_type,
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
        )



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from typing import Dict, Literal, Optional

import numpy as np
import onnx
from onnx.helper import tensor_dtype_to_np_dtype

from .evaluator import Evaluator
from .graph import OnnxGraph


def show_difference(
    a: np.ndarray, b: np.ndarray, atol: float = 1e-8, rtol: Optional[float] = None
):
    """Print the different element from array a and b."""
    if rtol is not None:
        mask = np.abs(1 - b / a) >= rtol
    mask = np.abs(a - b) >= atol
    return f"a={a[mask]}, b={b[mask]}"


def check_accuracy(
    model1: str | os.PathLike | onnx.ModelProto,
    model2: str | os.PathLike | onnx.ModelProto,
    input_maps: Optional[Dict[str, np.ndarray]] = None,
    backend: Literal["onnx", "onnxruntime", "openvino"] = "onnx",
) -> Dict[str, Dict[str, float]]:
    """
    Check the accuracy of two ONNX models.

    Args:
        model1 (onnx.ModelProto): The first ONNX model to be compared.
        model2 (onnx.ModelProto): The second ONNX model to be compared.
        input_maps (Dict[str, numpy.ndarray], optional): The input data to be used for
            the comparison. If not provided, the model will be run with random input
            data.
        backend (Literal["onnx", "onnxruntime", "openvino"], optional): The backend
            to be used for the comparison. Defaults to "onnx".

    Returns:
        Dict[str, Dict[str, float]]: A dictionary containing the accuracy metrics.
    """

    if not isinstance(model1, onnx.ModelProto):
        model1 = onnx.load_model(model1)
    if not isinstance(model2, onnx.ModelProto):
        model2 = onnx.load_model(model2)
    graph1 = OnnxGraph(model1)
    graph2 = OnnxGraph(model2)

    if input_maps is None:
        input_maps1 = {}
        input_maps2 = {}
        for input_name in graph1.inputs:
            assert input_name in graph2.inputs
            shape = graph1.static_tensor_shape(input_name)
            etype = graph1.tensor_type(input_name)
            dtype = tensor_dtype_to_np_dtype(etype)
            input_maps1[input_name] = np.asarray(np.random.randn(*shape)).astype(dtype)
            # data type may change
            etype = graph2.tensor_type(input_name)
            dtype = tensor_dtype_to_np_dtype(etype)
            input_maps2[input_name] = input_maps1[input_name].astype(dtype)
    else:
        input_maps1 = input_maps
        input_maps2 = input_maps
    output_maps = graph1.outputs
    for output_name in graph2.outputs:
        assert output_name in output_maps

    runner1 = Evaluator(model1, backend)
    runner2 = Evaluator(model2, backend)
    results1 = runner1(list(output_maps), input_maps1)
    results2 = runner2(list(output_maps), input_maps2)

    error_maps: Dict[str, Dict[str, float]] = {}
    for name, x, y in zip(output_maps, results1, results2):
        abs_error = np.abs(x - y)
        rel_mask = np.abs(x) > np.finfo(x.dtype).eps
        rel_error = np.abs(x - y)[rel_mask] / np.abs(x[rel_mask])
        error_maps[name] = {
            "ABS": abs_error.mean(),
            "REL": rel_error.mean(),
        }
    return error_maps



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ
from typing import Dict, List, Tuple, Union

import numpy as np
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register(name="inspect_sparsity_ratio", deps=["initializer_to_constant"])
class InspectSparsityRatio(Rewriter):
    """Inspect Sparsity Ratio"""

    _global_sparsity: Dict[int, Dict[str, Tuple[Tuple[int, ...], float]]] = {}
    """Global sparsity mapping, graph id to layer->(shape, sparsity ratio)"""

    def __init__(self):
        node = SingleNodePattern("Conv") | SingleNodePattern("ConvTranspose")
        super().__init__(node)
        self.register_post_hook(self._print_global_sparsity)

    def _print_global_sparsity(self, graph: OnnxGraph) -> OnnxGraph:
        sparse_map = self._global_sparsity.get(id(graph))
        if sparse_map:
            total_size: float = 0
            sparse_size: float = 0
            for shape, ratio in sparse_map.values():
                total_size += np.prod(shape, dtype="float")
                sparse_size += np.prod(shape, dtype="float") * ratio
            if total_size > (1 << 30):
                unit, unit_text = 1e9, "GiB"
            elif total_size > (1 << 20):
                unit, unit_text = 1e6, "MiB"
            elif total_size > (1 << 10):
                unit, unit_text = 1e3, "KiB"
            else:
                unit, unit_text = 1, "B"
            logger.info(f"Total size: {total_size / unit:.2f} {unit_text}")
            logger.info(f"Sparse size: {sparse_size / unit:.2f} {unit_text}")
            logger.info(f"Sparsity ratio: {sparse_size / total_size:.2%}")
        return graph

    def get_sparse_ratio(
        self,
        weight: NodeProto,
        zero_point: Union[int, np.ndarray] = 0,
        op_type: str = "Conv",
    ) -> Tuple[Tuple[int, ...], float]:
        """Inspect sparse ratio on specific location"""
        weight_value = self.get_value(weight)
        if weight_value is None:
            return (), -1

        # expand zero_point's shape
        if isinstance(zero_point, np.ndarray):
            new_shape = [1] * len(weight_value.shape)
            if op_type == "ConvTranspose":
                new_shape[1] = zero_point.size
            else:
                new_shape[0] = zero_point.size
            zero_point = zero_point.reshape(new_shape)
        sparse_ratio = np.count_nonzero(weight_value == zero_point) / weight_value.size
        return weight_value.shape, sparse_ratio

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        if id(graph) not in self._global_sparsity:
            self._global_sparsity[id(graph)] = {}
        sparse_map = self._global_sparsity[id(graph)]

        node = nodes[0]
        weight = self.get_input_node(node, 1)
        if weight is None:
            return

        sparse_ratio = -1.0
        if weight.op_type == "Constant":
            zero_point = 0
        elif weight.op_type == "DequantizeLinear":
            dequantize = weight
            # use DequantizeLinear's x as weight
            weight = self.get_input_node(dequantize, 0)
            if weight is None:
                return
            zero_point = self.get_value(dequantize.input[2])
            if zero_point is None:
                zero_point = 0
        else:
            return
        shape, sparse_ratio = self.get_sparse_ratio(weight, zero_point, node.op_type)

        if sparse_ratio == -1:
            return
        logger.info(f"{node.name}: {sparse_ratio:.2%}")
        sparse_map[node.name] = (shape, sparse_ratio)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info
from onnx.numpy_helper import from_array
from onnx.onnx_pb import TensorProto

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET
from onnxifier.checker import show_difference
from onnxifier.evaluator import Evaluator
from onnxifier.graph import OnnxGraph
from onnxifier.pass_manager import PassManager


def _make_conv_graph():
    cast = make_node(
        "Cast",
        inputs=["W"],
        outputs=["WF16"],
        to=TensorProto.FLOAT,
        name="cast",
    )
    mul = make_node(
        "Mul",
        inputs=["WF16", "scale"],
        outputs=["W_dequant"],
        name="mul",
    )
    dq = make_node(
        "DequantizeLinear",
        inputs=["X", "x_scale", "x_zp"],
        outputs=["X_dequant"],
        name="dq",
    )
    conv = make_node(
        "Conv",
        inputs=["X_dequant", "W_dequant"],
        outputs=["Y"],
        name="conv",
    )
    graph = make_graph(
        [dq, cast, mul, conv],
        name="test",
        inputs=[make_tensor_value_info("X", TensorProto.UINT8, [1, 8, 32, 32])],
        outputs=[make_tensor_value_info("Y", TensorProto.FLOAT, [1, 4, "H", "W"])],
        initializer=[
            from_array(np.random.uniform(-128, 127, [4, 8, 3, 3]).astype(np.int8), "W"),
            from_array(np.ones([4, 1, 1, 1], np.float32) / 127, "scale"),
            from_array(np.ones([8], np.float32) / 255, "x_scale"),
            from_array(np.zeros([8], np.uint8), "x_zp"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def _make_matmul_graph():
    cast = make_node(
        "Cast",
        inputs=["W"],
        outputs=["WF16"],
        to=TensorProto.FLOAT,
        name="cast",
    )
    mul = make_node(
        "Mul",
        inputs=["WF16", "scale"],
        outputs=["W_dequant"],
        name="mul",
    )
    gemm = make_node(
        "MatMul",
        inputs=["X", "W_dequant"],
        outputs=["Y"],
        name="gemm",
    )
    graph = make_graph(
        [cast, mul, gemm],
        name="test",
        inputs=[make_tensor_value_info("X", TensorProto.FLOAT, [1, 8, 16, 32])],
        outputs=[make_tensor_value_info("Y", TensorProto.FLOAT, [1, "C", "H", "W"])],
        initializer=[
            from_array(np.random.uniform(-128, 127, [8, 32, 16]).astype(np.int8), "W"),
            from_array(np.ones([8, 1, 1], np.float32), "scale"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def _make_gemm_graph():
    cast = make_node(
        "Cast",
        inputs=["W"],
        outputs=["WF16"],
        to=TensorProto.FLOAT,
        name="cast",
    )
    mul = make_node(
        "Mul",
        inputs=["WF16", "scale"],
        outputs=["W_dequant"],
        name="mul",
    )
    gemm = make_node(
        "Gemm",
        inputs=["X", "W_dequant"],
        outputs=["Y"],
        name="gemm",
    )
    graph = make_graph(
        [cast, mul, gemm],
        name="test",
        inputs=[make_tensor_value_info("X", TensorProto.FLOAT, [8, 16])],
        outputs=[make_tensor_value_info("Y", TensorProto.FLOAT, ["M", "N"])],
        initializer=[
            from_array(np.random.uniform(-128, 127, [16, 8]).astype(np.int8), "W"),
            from_array(np.ones([16, 1], np.float32), "scale"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def test_conv_dequantize_weight():
    graph = _make_conv_graph()
    runner1 = Evaluator(graph.model)
    pm = PassManager(["conv_dequantize_weight"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    x = np.random.randint(0, 255, size=[1, 8, 32, 32], dtype=np.uint8)
    y1 = runner1(["Y"], {"X": x})[0]
    y2 = runner2(["Y"], {"X": x})[0]

    assert np.allclose(y1, y2, rtol=1e-2), show_difference(y1, y2, rtol=1e-2)


def test_matmul_dequantize_weight():
    graph = _make_matmul_graph()
    runner1 = Evaluator(graph.model)
    pm = PassManager(["gemm_dequantize_weight"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    x = np.random.rand(1, 8, 16, 32).astype(np.float32)
    y1 = runner1(["Y"], {"X": x})[0]
    y2 = runner2(["Y"], {"X": x})[0]

    assert np.allclose(y1, y2, rtol=1e-6), show_difference(y1, y2, rtol=1e-6)


def test_gemm_dequantize_weight():
    graph = _make_gemm_graph()
    runner1 = Evaluator(graph.model)
    pm = PassManager(["gemm_dequantize_weight"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    x = np.random.rand(8, 16).astype(np.float32)
    y1 = runner1(["Y"], {"X": x})[0]
    y2 = runner2(["Y"], {"X": x})[0]

    assert np.allclose(y1, y2, rtol=1e-6), show_difference(y1, y2, rtol=1e-6)


def test_recalculate_dequantize_weight():
    graph = _make_conv_graph()
    runner1 = Evaluator(graph.model)
    pm = PassManager(["recalculate_dequantize_weight"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    x = np.random.rand(1, 8, 32, 32).astype(np.uint8)
    y1 = runner1(["Y"], {"X": x})[0]
    y2 = runner2(["Y"], {"X": x})[0]

    assert np.allclose(y1, y2, rtol=1e-6), show_difference(y1, y2, rtol=1e-6)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from onnx.helper import make_tensor_type_proto, make_value_info
from onnx.numpy_helper import from_array, to_array

from ... import OnnxGraph
from .. import PASSES
from ..utils import make_constant


@PASSES.register()
def eliminate_initializer_input(graph: OnnxGraph) -> OnnxGraph:
    """Eliminate graph input that is actually an initializer"""
    pop_list = []
    for init in graph.initializer:
        if init.name in graph.inputs:
            i = graph.inputs.pop(init.name)
            pop_list.append(i)
    for i in sorted(pop_list, reverse=True):
        graph.input.pop(i)

    # re-assign index of inputs.
    for i, node in enumerate(graph.input):
        graph.inputs[node.name] = i
    return graph


@PASSES.register(deps=["initializer_unique", "eliminate_initializer_input"])
def initializer_to_constant(graph: OnnxGraph) -> OnnxGraph:
    """Convert initializer value to node Constant."""
    for init in graph.initializer:
        node = make_constant(init.name, to_array(init))
        if node.name in graph:
            # initializer could have same name as other node,
            # rename it to avoid conflict.
            node.name += "_init"
        node.output[0] = init.name
        graph._value_info.append(  # pylint: disable=W0212
            make_value_info(
                node.output[0],
                make_tensor_type_proto(init.data_type, init.dims),
            )
        )
        graph.add_onnx_node(node)
    while graph.initializer:
        graph.initializer.pop()
    return graph


@PASSES.register(name="eliminate_duplicated_initializer")
def eliminate_duplicated_initializer(graph: OnnxGraph) -> OnnxGraph:
    """Eliminate duplicated initializer.
    It can be used for initializer_to_constant"""
    init_set = {init.name: init for init in graph.initializer}
    while graph.initializer:
        graph.initializer.pop()
    graph.initializer.extend(init_set.values())
    return graph


@PASSES.register(name="initializer_unique")
def initializer_unique(graph: OnnxGraph) -> OnnxGraph:
    """Makes all initializer unique but rarely needed.
    Generally only eliminate_duplicated_initializer is needed.
    """
    init_name_map = {}
    for init in graph.initializer:
        init_name_map[init.name] = {
            "init": init,
            "times": 0,
        }

    init_add = []
    for node in graph:
        node_pb = graph.nodes[node]["pb"]
        for i, name in enumerate(node_pb.input):
            if name in init_name_map:
                init_name_map[name]["times"] += 1
                if init_name_map[name]["times"] == 1:
                    continue
                new_name = name + f"/{init_name_map[name]['times']}"
                node_pb.input[i] = new_name
                new_init = from_array(to_array(init_name_map[name]["init"]), new_name)
                init_add.append(new_init)

    # should remove duplicated items in initializer
    graph._model.graph.ClearField("initializer")  # pylint: disable=W0212
    graph.initializer.extend(init_add)
    for v in init_name_map.values():
        graph.initializer.append(v["init"])
    return graph



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from contextlib import suppress

import numpy as np
import onnx
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from .. import OnnxGraph, cast_in, make_constant
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="interpolate")
class Interpolate(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/image/interpolate-11.html

    https://onnx.ai/onnx/operators/onnx__Resize.html
    """

    def _to_mode(self, mode):
        if mode == "linear_onnx":
            return "linear"
        return mode

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        shape_calculation_mode = self.get_attribute(ori_node, "shape_calculation_mode")
        if shape_calculation_mode is None:
            # for opset1 compatibility
            return self.replace_opset1(graph, ori_node)
        # axes
        version = self.get_attribute(ori_node, "version")
        if len(ori_node.input) > 3:
            axes = self.get_value(ori_node.input[3])
            ori_node.input.pop(3)
        elif version == "opset11":
            # axes is optional
            if len(ori_node.input) > 2:
                axes = self.get_value(ori_node.input[2])
            else:
                image_shape = graph.tensor_shape(ori_node.input[0])
                axes = np.arange(len(image_shape))
        else:
            rank = graph.tensor_shape(ori_node.input[1])[0]
            if isinstance(rank, str) or rank < 1:
                raise ValueError(f"Input shape[0] is dynamic: {rank}")
            # NOTE: 1. negative axis causes a bug in onnxruntime
            # NOTE: 2. axis less than input rank causes a bug in openvino
            # Solution to #1: convert to positive axis (here line:64)
            # Solution to #2: use pass "resize_remove_axes"
            axes = np.arange(-rank, 0) + len(graph.tensor_shape(ori_node.input[0]))
        self.set_attribute(ori_node, "axes", axes)
        while len(ori_node.input) < 3:
            ori_node.input.append("")
        if shape_calculation_mode == "sizes":
            ori_node.input.append(ori_node.input[1])
            ori_node.input[1] = ori_node.input[2] = ""
            size_type = graph.tensor_type(ori_node.input[3])
            if size_type != onnx.TensorProto.INT64:
                self += cast_in(ori_node, 3, onnx.TensorProto.INT64)
        elif shape_calculation_mode == "scales":
            if version == "opset11":
                ori_node.input[2] = ori_node.input[1]
            ori_node.input[1] = ""

        mode = self.get_attribute(ori_node, "mode") or "nearest"
        ctm = self.get_attribute(ori_node, "coordinate_transformation_mode")
        nearest_mode = self.get_attribute(ori_node, "nearest_mode")
        cubic_coeff_a = self.get_attribute(ori_node, "cubic_coeff")
        pads_begin = self.get_attribute(ori_node, "pads_begin")
        pads_begin = list(map(int, pads_begin.split(",")))  # type: ignore
        pads_end = self.get_attribute(ori_node, "pads_end")
        pads_end = list(map(int, pads_end.split(",")))  # type: ignore

        if any(pads_begin) or any(pads_end):
            raise ValueError("Interpolate does not support converting with padding")

        return make_node(
            "Resize",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
            mode=self._to_mode(mode),
            coordinate_transformation_mode=ctm or "half_pixel",
            nearest_mode=nearest_mode or "round_prefer_floor",
            cubic_coeff_a=cubic_coeff_a or -0.75,
            axes=axes,
        )

    def replace_opset1(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        """Interpolate opset1 doesn't have attribute "shape_calculation_mode", we
        have to guess the scales_or_size based on its data type.
        """
        # get scales_or_size
        scales_or_size = self.get_value_or_die(ori_node.input[1])
        # size to scales
        input_dims = np.array(graph.tensor_shape(ori_node.input[0]))
        scales = scales_or_size
        with suppress(ValueError):
            np.iinfo(scales_or_size.dtype)  # integers for size
            if scales.size < input_dims.size:
                scales = np.concatenate(
                    [input_dims[: input_dims.size - scales.size], scales]
                )
            scales = np.true_divide(scales, input_dims).astype("float32")
        if scales.size < input_dims.size:
            preleading_ones = np.ones_like(input_dims[: input_dims.size - scales.size])
            scales = np.concatenate([preleading_ones, scales]).astype("float32")
        # make empty roi
        ori_node.input[1] = ""
        # make scales
        scale_node = make_constant(f"{ori_node.name}/scales", scales)
        if len(ori_node.input) > 2:
            ori_node.input[2] = scale_node.output[0]
        else:
            ori_node.input.append(scale_node.output[0])
        self += scale_node
        # axes
        if len(ori_node.input) > 3:
            axes = self.get_value_or_die(ori_node.input[3])
            assert tuple(axes) == (2, 3)
            ori_node.input.pop(3)

        mode = self.get_attribute(ori_node, "mode") or "nearest"
        ctm = self.get_attribute(ori_node, "coordinate_transformation_mode")
        nearest_mode = self.get_attribute(ori_node, "nearest_mode")
        cubic_coeff_a = self.get_attribute(ori_node, "cubic_coeff")
        pads_begin = self.get_attribute(ori_node, "pads_begin")
        pads_begin = list(map(int, pads_begin.split(",")))  # type: ignore
        pads_end = self.get_attribute(ori_node, "pads_end")
        pads_end = list(map(int, pads_end.split(",")))  # type: ignore

        if any(pads_begin) or any(pads_end):
            raise ValueError("Interpolate does not support converting with padding")

        return make_node(
            "Resize",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
            mode=self._to_mode(mode),
            coordinate_transformation_mode=ctm or "half_pixel",
            nearest_mode=nearest_mode or "round_prefer_floor",
            cubic_coeff_a=cubic_coeff_a or -0.75,
        )



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Remove conv weights with all zero channels
"""

from itertools import chain
from typing import Dict, List, Protocol, Set

import networkx as nx
import numpy as np
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from ...algo.subgraph import find_inbound_nodes, find_outbound_nodes
from .. import PASSES, Registry
from ..pattern import Pattern, SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant

_CHAN_MIXER = ("Conv", "ConvTranspose")
_WHITELIST = (
    "Cast",
    "Clip",
    "Erf",
    "Gelu",
    "HardSigmoid",
    "HardSwish",
    "LeakyRelu",
    "MaxPool",
    "Mish",
    "PRelu",
    "Relu",
    "Sigmoid",
    "Swish",
    "Tanh",
)
_BLACKLIST = ()


class PrunedData:
    """Helper class to store pruned data."""

    def __init__(self):
        # mapping node to its output channel indices of zero values
        # E.g:
        # Y = Conv(X, W), Y[:, ind] == 0
        # out_zero_indices[Y] = {ind}
        self.fw_zero_indices: Dict[NodeProto, Set[int]] = {}
        # mapping node's predecessor to this node's input channel indices of zero values
        # E.g:
        # Y = Conv(X, W), X[:, ind] == 0
        # in_zero_indices[X] = {ind}
        # Note: input channels will be merged if two nodes share the same input
        self.bw_zero_indices: Dict[NodeProto, Set[int]] = {}
        self.bw_stopper: Dict[NodeProto, bool] = {}
        self.axes_map: Dict[NodeProto, Set[int]] = {}
        self.seen_nodes: Set[str] = set()


class IPruneForwardFunc(Protocol):
    """Interface for forward pruning functions."""

    __DEPS__: List[str]
    __PATCHES__: List[str]

    def __call__(
        self, rewriter: Rewriter, node: NodeProto, pruned: PrunedData
    ) -> Set[int]:
        return set()


class IPruneBackwardFunc(Protocol):
    """Interface for backward pruning functions."""

    __DEPS__: List[str]
    __PATCHES__: List[str]

    def __call__(
        self, rewriter: Rewriter, node: NodeProto, pruned: PrunedData
    ) -> List[Set[int]]:
        return []


FW_PRUNE = Registry[IPruneForwardFunc]("ForwardPruner")
BW_PRUNE = Registry[IPruneBackwardFunc]("BackwardPruner")
MIXTURE = Registry[IPruneForwardFunc]("Mixture")


def _pass_through_forward(
    self: Rewriter,
    node: NodeProto,
    pruned: PrunedData,
):
    pred = self.graph.onnx_predecessors(node)[0]
    return pruned.fw_zero_indices[pred]


def _pass_through_backward(
    self: Rewriter,  # pylint: disable=unused-argument
    node: NodeProto,
    pruned: PrunedData,
):
    return [pruned.bw_zero_indices[node]]


def _default_mixture(
    self: Rewriter,  # pylint: disable=unused-argument
    node: NodeProto,
    pruned: PrunedData,
):
    return pruned.fw_zero_indices[node]


for op_type in _WHITELIST:
    FW_PRUNE.register(op_type)(_pass_through_forward)
    BW_PRUNE.register(op_type)(_pass_through_backward)


def _get_weights(rewriter: Rewriter, node: NodeProto) -> np.ndarray:
    if node.op_type in ("Conv", "ConvTranspose"):
        weights = rewriter.get_value(node.input[1])
    else:
        raise ValueError(f"Unsupported op_type: {node.op_type}")

    if weights is None:
        raise ValueError(f"Non constant weights of node: {node.name}")
    return weights


def _get_zero_indices(value: np.ndarray, axis: int = 0) -> Set[int]:
    zero_indices = set()
    for i, val in enumerate(np.split(value, value.shape[axis], axis=axis)):
        if np.all(val == 0):
            zero_indices.add(i)
    return zero_indices


@FW_PRUNE.register("Conv")
def get_conv_zero_out_channels(
    self: Rewriter,
    node: NodeProto,
    pruned: PrunedData,  # pylint: disable=unused-argument
) -> Set[int]:
    """Get zeroed output channels of a Conv operator.

    A zeroed output channel ``i`` is a channel that all weights of ``W[i][:]``
    are zeros. If the Conv has bias, the corresponding bias value must be zero
    too.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the Conv operator to be checked
        pruned (PrunedData): a helper dictionary to store the pruned channels
    """
    weights = _get_weights(self, node)
    if group := self.get_attribute(node, "group"):
        if group and group != 1 and weights.shape[1] != 1:
            # not a depthwise conv but with group > 1
            return set()

    oi_w = np.abs(weights).sum(axis=(-1, -2))
    zero_indices = _get_zero_indices(oi_w)
    if len(node.input) > 2:
        bias = self.get_value(node.input[2])
        if bias is None:
            raise ValueError(f"Cannot get a constant value of bias: {node.name}")
        zero_indices.intersection_update(_get_zero_indices(bias))
    return zero_indices


@FW_PRUNE.register("ConvTranspose")
def get_convtranspose_zero_out_channels(
    self: Rewriter,
    node: NodeProto,
    pruned: PrunedData,  # pylint: disable=unused-argument
) -> Set[int]:
    """Get zeroed output channels of a ConvTranspose operator.

    A zeroed output channel ``i`` is a channel that all weights of ``W[:][i]``
    are zeros. If the ConvTranspose has bias, the corresponding bias value must
    be zero too.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the ConvTranspose operator to be checked
        pruned (PrunedData): a helper dictionary to store the pruned channels
    """
    weights = _get_weights(self, node)
    if group := self.get_attribute(node, "group"):
        if group and group != 1 and weights.shape[1] != 1:
            # not a depthwise deconv but with group > 1
            return set()

    io_w = np.abs(weights).sum(axis=(-1, -2))
    zero_indices = _get_zero_indices(io_w, axis=1)
    if len(node.input) > 2:
        bias = self.get_value(node.input[2])
        if bias is None:
            raise ValueError(f"Cannot get a constant value of bias: {node.name}")
        zero_indices.intersection_update(_get_zero_indices(bias))
    return zero_indices


@FW_PRUNE.register("Constant")
def get_constant_zero_out_channels(
    self: Rewriter,
    node: NodeProto,
    pruned: PrunedData,  # pylint: disable=unused-argument
) -> Set[int]:
    """Get zeroed output channels of a constant operator.

    A zeroed output channel ``i`` is a channel that all values are zeros.

    Note:
        The channel axis is fixed to 1.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the Constant operator to be checked
        pruned (PrunedData): unused parameter
    """
    value = self.get_value(node)
    assert value is not None
    zero_indices = _get_zero_indices(value, axis=1)
    return zero_indices


# not ready
def get_add_zero_out_channels(
    self: Rewriter, node: NodeProto, pruned: PrunedData
) -> Set[int]:
    """Get zeroed output channels of an Add or Sub operator.

    A zeroed output channel ``i`` is a channel that all values from both
    operands are zeros.

    Note:
        The channel axis is fixed to 1.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the Add/Sub operator to be checked
        pruned (dict): a helper dictionary to store the pruned channels
    """
    operand_0 = self.get_input_node(node, 0)
    operand_1 = self.get_input_node(node, 1)
    if operand_0 is None:
        value = self.get_value(node.input[0])
        assert value is not None
        indices_0 = _get_zero_indices(value, axis=1)
        axis_0 = value.shape[1]
    else:
        indices_0 = pruned.fw_zero_indices[operand_0]
        axis_0 = self.graph.tensor_shape(node.input[0])[1]
    if operand_1 is None:
        value = self.get_value(node.input[1])
        assert value is not None
        indices_1 = _get_zero_indices(value, axis=1)
        axis_1 = value.shape[1]
    else:
        indices_1 = pruned.fw_zero_indices[operand_1]
        axis_1 = self.graph.tensor_shape(node.input[1])[1]

    if axis_0 == axis_1:
        zero_indices = indices_0.intersection(indices_1)
    elif axis_0 == 1:
        zero_indices = indices_1 if indices_0 else set()
    elif axis_1 == 1:
        zero_indices = indices_0 if indices_1 else set()
    else:
        raise ValueError(f"Broadcast error executing node: {node.name}")
    return zero_indices


@FW_PRUNE.register("Concat")
def get_concat_zero_out_channels(
    self: Rewriter, node: NodeProto, pruned: PrunedData
) -> Set[int]:
    """Get zeroed output channels of a Concat operator.

    A zeroed output channel ``i`` is a channel that corresponding values from one
    input are all zeros.

    Note:
        The channel axis is fixed to 1.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the Concat operator to be checked
        pruned (PrunedData): a helper dictionary to store the pruned channels
    """
    indices_list = []
    ch_axis = 1
    channels = 0
    for i, inp in enumerate(node.input):
        operand = self.get_input_node(node, i)
        if operand is None:
            value = self.get_value(node.input[i])
            assert value is not None
            indices = _get_zero_indices(value, axis=1)
        else:
            indices = pruned.fw_zero_indices[operand]
        indices_list.extend([i + channels for i in indices])
        shape = self.graph.tensor_shape(inp)[ch_axis]
        assert isinstance(shape, int)
        channels += shape
    zero_indices = set(indices_list)
    return zero_indices


@BW_PRUNE.register("Conv")
def get_conv_zero_in_channels(
    self: Rewriter,
    node: NodeProto,
    pruned: PrunedData,  # pylint: disable=unused-argument
) -> List[Set[int]]:
    """Get zeroed input channels of a Conv operator.

    A zeroed input channel ``i`` is a channel that all weights of ``W[:,i]`` are
    zeros.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the Conv operator to be checked
        pruned (PrunedData): unused parameter
    """
    weights = _get_weights(self, node)
    if group := self.get_attribute(node, "group"):
        if group and group != 1 and weights.shape[1] != 1:
            return [set()]

    oi_w = np.abs(weights).sum(axis=(-1, -2))
    zero_indices = _get_zero_indices(oi_w, axis=1)
    return [zero_indices]


@BW_PRUNE.register("ConvTranspose")
def get_convtranspose_zero_in_channels(
    self: Rewriter,
    node: NodeProto,
    pruned: PrunedData,  # pylint: disable=unused-argument
) -> List[Set[int]]:
    """Get zeroed input channels of a ConvTranspose operator.

    A zeroed input channel ``i`` is a channel that all weights of ``W[i,:]`` are
    zeros.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the ConvTranspose operator to be checked
        pruned (PrunedData): unused parameter
    """
    weights = _get_weights(self, node)
    if group := self.get_attribute(node, "group"):
        if group and group != 1 and weights.shape[1] != 1:
            return [set()]

    io_w = np.abs(weights).sum(axis=(-1, -2))
    zero_indices = _get_zero_indices(io_w, axis=0)
    return [zero_indices]


@BW_PRUNE.register("Concat")
def get_concat_zero_in_channels(
    self: Rewriter, node: NodeProto, pruned: PrunedData
) -> List[Set[int]]:
    """Get zeroed input channels of a Concat operator.

    A zeroed input channel ``i`` is a channel that corresponding values from
    output are all zeros.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the Concat operator to be checked
        pruned (PrunedData): a helper dictionary to store the pruned channels
    """
    axes_out = np.asarray(tuple(pruned.bw_zero_indices[node]))
    ch_axis = 1
    zero_indices = []
    for _, inp in enumerate(node.input):
        channels = self.graph.tensor_shape(inp)[ch_axis]
        assert isinstance(channels, int)
        zero_indices.append(set(axes_out[axes_out < channels]))
        axes_out = axes_out[axes_out >= channels] - channels
    return zero_indices


# not ready
def get_add_zero_in_channels(
    self: Rewriter, node: NodeProto, pruned: PrunedData
) -> List[Set[int]]:
    """Get zeroed input channels of an Add or Sub operator.

    A zeroed input channel ``i`` is a channel that corresponding values from both
    operands are all zeros.

    Args:
        self (Rewriter): used in class object :class:`RemoveZeroChannelsRewriter`
        node (NodeProto): the Add/Sub operator to be checked
        pruned (PrunedData): a helper dictionary to store the pruned channels
    """
    axes_out = pruned.bw_zero_indices[node]
    ch_axis = 1
    shape0 = self.graph.tensor_shape(node.input[0])
    shape1 = self.graph.tensor_shape(node.input[1])
    zero_indices: List[Set[int]] = [set(), set()]
    if len(shape0) > ch_axis and (ch := shape0[ch_axis]):
        if isinstance(ch, int) and ch > 1:
            zero_indices[0] = axes_out
    if len(shape1) > ch_axis and (ch := shape1[ch_axis]):
        if isinstance(ch, int) and ch > 1:
            zero_indices[1] = axes_out
    return zero_indices


@MIXTURE.register("Conv")
@MIXTURE.register("ConvTranspose")
def get_conv_mixture(
    self: Rewriter,  # pylint: disable=unused-argument
    node: NodeProto,
    pruned: PrunedData,
) -> Set[int]:
    """Get the mix result of zero channel indices consider both forward and backward
    paths.
    """
    if pruned.bw_stopper.get(node):
        return set()
    # could be output so bw_zero_indices could be empty
    return pruned.fw_zero_indices[node] | pruned.bw_zero_indices.get(node, set())


@PASSES.register("remove_zero_channels", deps=["onnxsim"])
class RemoveZeroChannelsRewriter(Rewriter):
    """Prune zerod channels in constant weights of Conv and ConvTranspose operators."""

    # global temp memory to propagate pruned axes through entire graph
    _memo: Dict[int, PrunedData] = {}

    def __init__(self):
        pattern = sum([SingleNodePattern(op) for op in _CHAN_MIXER])
        assert isinstance(pattern, Pattern)
        super().__init__(pattern=pattern)

    def _find_dep_graph(self, graph: OnnxGraph, node: NodeProto):
        nodes_found = {node}  # type: ignore
        searching_nodes = [(node, "up")]
        while searching_nodes:
            curr_node, direction = searching_nodes.pop(0)
            if direction == "down":
                nodes, endpoints = find_outbound_nodes(graph, curr_node, _CHAN_MIXER)
            else:  # up
                nodes, endpoints = find_inbound_nodes(graph, curr_node, _CHAN_MIXER)
            if new_nodes := (set(endpoints) - nodes_found.intersection(endpoints)):
                direction = "down" if direction == "up" else "up"  # revert direction
                searching_nodes.extend([(op, direction) for op in new_nodes])
            nodes_found.update(nodes + endpoints)
        return tuple(nodes_found)

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        if id(graph) not in self._memo:
            self._memo[id(graph)] = PrunedData()
        memo = self._memo[id(graph)]

        # find conv-conv subgraph
        boundary_nodes = self._find_dep_graph(graph, nodes[0])
        if len(boundary_nodes) == 1:
            return
        if any(op.op_type in _BLACKLIST for op in boundary_nodes):
            return

        logger.debug(
            f"Boundary around {nodes[0].name}: {[op.name for op in boundary_nodes]}"
        )
        h = graph.onnx_subgraph(boundary_nodes)
        start_points: List[NodeProto] = []
        for node_name, ind in h.in_degree():  # type: ignore
            if ind == 0 and h.nodes[node_name]["pb"].op_type in _CHAN_MIXER:
                start_points.append(h.nodes[node_name]["pb"])
        logger.debug(f"Start points: {[op.name for op in start_points]}")

        end_points: List[NodeProto] = []
        for node_name, outd in h.out_degree():  # type: ignore
            if outd == 0 and h.nodes[node_name]["pb"].op_type in _CHAN_MIXER:
                end_points.append(h.nodes[node_name]["pb"])
        logger.debug(f"End points: {[op.name for op in end_points]}")

        # get pruned axes
        for n in nx.topological_sort(h):
            self.update_zero_axes_fw(h.nodes[n]["pb"], memo)
        for n in nx.topological_sort(nx.reverse(h, copy=True)):  # BP
            self.update_zero_axes_bw(h.nodes[n]["pb"], memo)
        for i in nx.topological_sort(h):
            refresh_axes_out = self.update_pruned_axes(h.nodes[i]["pb"], memo)
            if refresh_axes_out:
                # refresh all downstreaming nodes of i
                for j in nx.topological_sort(nx.subgraph(h, nx.descendants(h, i))):
                    self.update_zero_axes_fw(h.nodes[j]["pb"], memo)

        for op in chain(start_points, end_points):
            self.remove_weights_channels(op, memo)

    def update_zero_axes_fw(self, node: NodeProto, pruned: PrunedData):
        """Get pruned axis of node in topological order and store in
        :class:`PrunedData`.
        """

        if get_zero_channels := FW_PRUNE.get(node.op_type):
            zero_indices = get_zero_channels(self, node, pruned)
            pruned.fw_zero_indices[node] = zero_indices
            logger.debug(f"zero out channels of {node.name}: {zero_indices}")
        else:
            pruned.fw_zero_indices[node] = set()

    def update_zero_axes_bw(self, node: NodeProto, pruned: PrunedData):
        """Get pruned axis of node in reverse topological order and store in
        :class:`PrunedData`.
        """

        if get_zero_channels := BW_PRUNE.get(node.op_type):
            indices_of_inputs = get_zero_channels(self, node, pruned)
            stopper = False
        else:
            indices_of_inputs = [set()] * len(node.input)
            stopper = True
        for pred, ind in zip(self.graph.onnx_predecessors(node), indices_of_inputs):
            if pred not in pruned.bw_zero_indices:
                pruned.bw_zero_indices[pred] = ind
            else:
                # if two nodes go backward to a same output port
                # the available zero channels should be the intersection of both
                pruned.bw_zero_indices[pred] &= ind
            if stopper:
                pruned.bw_stopper[pred] = True
        logger.debug(f"zero in channels of {node.name}: {indices_of_inputs}")

    def update_pruned_axes(self, node: NodeProto, memo: PrunedData) -> bool:
        """Calculate the final pruned axes by mixing the results from upstream and
        sibling nodes.
        """
        is_output_node = self.graph.out_degree(node.name) == 0
        need_update = is_output_node and node in memo.fw_zero_indices
        need_update |= node in memo.bw_zero_indices and node in memo.fw_zero_indices
        need_refresh = False
        if need_update:
            # need a mix
            if mix_channels := MIXTURE.get(node.op_type):
                out_renew_axes = mix_channels(self, node, memo)
            else:
                out_renew_axes = _default_mixture(self, node, memo)
            indices = set() if is_output_node else out_renew_axes
            need_refresh = memo.fw_zero_indices[node] != indices
            memo.fw_zero_indices[node] = memo.axes_map[node] = indices
            logger.debug(f"pruned out channels of {node.name}: {indices}")
        return need_refresh

    def remove_weights_channels(self, node: NodeProto, memo: PrunedData):
        """Remove zero output channels in weights."""

        if node.name in memo.seen_nodes:
            return

        out_axes = memo.axes_map.get(node)
        producer = self.get_input_node(node, 0)
        if out_axes is None:
            logger.debug(f"axes map of {node.name} is empty")
            return
        if producer is None:
            in_axes: Set[int] | None = set()
        else:
            in_axes = memo.axes_map.get(producer)
        if in_axes is None:
            logger.debug(f"input axes of {node.name} is empty")
            return
        elif not in_axes and not out_axes:
            logger.debug(f"not a sparse op: {node.name}")
            memo.seen_nodes.add(node.name)
            return  # not a sparse op

        weights = _get_weights(self, node)
        out_channel_axis = 0
        in_channel_axis = 1
        if node.op_type in ("ConvTranspose",):
            out_channel_axis = 1
            in_channel_axis = 0
        kept0 = set(range(weights.shape[out_channel_axis])) - out_axes
        kept1 = set(range(weights.shape[in_channel_axis])) - in_axes
        pruned_weights = np.take(weights, tuple(kept0), axis=out_channel_axis)
        pruned_weights = np.take(pruned_weights, tuple(kept1), axis=in_channel_axis)
        weight_node = make_constant(f"{node.name}/pruned_weight", pruned_weights)
        self += weight_node
        node.input[1] = weight_node.output[0]

        if len(node.input) > 2:
            # update bias
            bias = self.get_value(node.input[2])
            if bias is None:
                raise ValueError(f"Cannot get a constant value of bias: {node.name}")
            pruned_bias = np.take(bias, tuple(kept0))
            bias_node = make_constant(f"{node.name}/pruned_bias", pruned_bias)
            self += bias_node
            node.input[2] = bias_node.output[0]
        self += node
        memo.seen_nodes.add(node.name)
        logger.debug(
            f"update weights of {node.name}: {weights.shape}->{pruned_weights.shape}"
        )



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ...graph import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register(name="merge_parallel_conv", deps=["initializer_to_constant"])
class MergeParallelConvRewrite(Rewriter):
    """Merge parallel conv ops into one conv op"""

    def __init__(self):
        pattern = GraphPattern()
        conv1 = SingleNodePattern("Conv")
        conv2 = SingleNodePattern("Conv")
        concat = SingleNodePattern("Concat")

        pattern.add_edge(conv1, concat)
        pattern.add_edge(conv2, concat)
        super().__init__(pattern)

    def make_concat(self, input_name_list: List[str], name: str, axis: int = 0):
        """concat two input"""
        concat = make_node(
            op_type="Concat",
            inputs=input_name_list[:],
            outputs=[f"{name}/Output"],
            name=name,
            axis=axis,
        )
        return concat

    def assign_nodes(self, nodes: List[NodeProto]):
        """Assign nodes by op_type, we find the order is not topo"""
        conv1, conv2, concat = None, None, None

        # assign concat
        for node in nodes:
            if node.op_type == "Concat":
                concat = node
        assert concat

        # assign conv by concat input
        conv_outputs = {
            node.output[0]: node for node in nodes if node.op_type == "Conv"
        }
        for name in concat.input:
            if name in conv_outputs:
                if conv1 is None:
                    conv1 = conv_outputs[name]
                else:
                    conv2 = conv_outputs[name]

        assert conv1
        assert conv2

        return conv1, conv2, concat

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        if len(nodes) != 3:
            return
        conv1, conv2, concat = self.assign_nodes(nodes)

        # TODO: check conv attrs
        # check continuous
        conv1_pos = [
            i for i, name in enumerate(concat.input) if name == conv1.output[0]
        ][0]
        conv2_pos = [
            i for i, name in enumerate(concat.input) if name == conv2.output[0]
        ][0]
        if conv1_pos + 1 != conv2_pos:
            return

        # check same input
        if conv1.input[0] != conv2.input[0]:
            return

        # both have bias or not
        if len(conv1.input) != len(conv2.input):
            return

        # only support channel concat
        if self.get_attribute(concat, "axis") != 1:
            return

        # concat weight on out channel(axis=0)
        w_node = self.make_concat(
            [conv1.input[1], conv2.input[1]], name=f"{conv1.input[1]}/MergedWeight"
        )
        conv1.input[1] = w_node.output[0]
        self += w_node

        # concat bias
        if len(conv1.input) == 3:
            b_node = self.make_concat(
                [conv1.input[2], conv2.input[2]], name=f"{conv1.input[2]}/MergedBias"
            )
            conv1.input[2] = b_node.output[0]
            self += b_node

        # pop conv2
        concat.input.pop(conv2_pos)

        self -= [conv2]



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-docstring

from typing import List

import numpy as np
import onnx
import pytest
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info
from onnx.onnx_pb import NodeProto

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET
from onnxifier.graph import OnnxGraph
from onnxifier.passes.pattern import GraphPattern, Pattern, SingleNodePattern
from onnxifier.passes.rewriter import Rewriter
from onnxifier.passes.utils import make_constant


class TestRewriter(Rewriter):
    def __init__(self, pattern, repeat, test_func):
        super().__init__(pattern, repeat)
        self.test_func = test_func

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        self.test_func(self, graph, nodes, *args, **kwargs)


def _make_graph(n_nodes=1):
    node = onnx.helper.make_node(
        "Dummy",
        ["x"],
        ["y"],
        "dummy",
        domain="test",
        attr1=1.0,
        attr2=2,
        attr3="3",
        attr4=onnx.numpy_helper.from_array(np.array([4])),
        attr6=onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, [6, 6, 6]),
        attr7=[7.0, 7.0, 7.0],
        attr8=[8, 8, 8],
        attr9=["9", "9", "9"],
        attr10=[onnx.numpy_helper.from_array(np.array([10]))],
        attr12=[
            onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, [12, 12, 12]),
            onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, [12, 12, 12]),
        ],
    )
    nodes = [node]
    for i in range(1, n_nodes):
        nodes[-1].output[0] = f"n{i}"
        nodes.append(
            onnx.helper.make_node(
                "DummyTwo",
                [f"n{i}"],
                ["y"],
                f"dummy{i}",
                domain="test",
            )
        )
    graph = onnx.helper.make_graph(
        nodes,
        "test",
        [onnx.helper.make_tensor_value_info("x", onnx.TensorProto.FLOAT, [1, 2, 3])],
        [onnx.helper.make_tensor_value_info("y", onnx.TensorProto.FLOAT, [1, 2, 3])],
    )
    return OnnxGraph(onnx.helper.make_model(graph))


def test_access_attribute():
    passed = False

    def _func(self, graph, nodes, *args, **kwargs):
        node = nodes[0]
        assert self.get_attribute(node, "attr1") == 1.0
        assert self.get_attribute(node, "attr2") == 2
        assert self.get_attribute(node, "attr3") == "3"
        assert (self.get_attribute(node, "attr4") == np.array([4])).all()
        assert isinstance(self.get_attribute(node, "attr6"), onnx.TypeProto)
        assert self.get_attribute(node, "attr7") == [7.0, 7.0, 7.0]
        assert self.get_attribute(node, "attr8") == [8, 8, 8]
        assert self.get_attribute(node, "attr9") == ["9", "9", "9"]
        assert (self.get_attribute(node, "attr10")[0] == np.array([10])).all()
        assert isinstance(self.get_attribute(node, "attr12")[0], onnx.TypeProto)
        assert isinstance(self.get_attribute(node, "attr12")[1], onnx.TypeProto)
        self.set_attribute(node, "attr1", 10.0)
        self.set_attribute(node, "attr2", 20)
        self.set_attribute(node, "attr3", "30")
        self.set_attribute(node, "attr4", np.array([40.0]))
        self.set_attribute(
            node,
            "attr6",
            onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, [60, 60, 60]),
        )
        self.set_attribute(node, "attr7", [70.0, 70.0, 70.0])
        self.set_attribute(node, "attr8", [80, 80, 80])
        self.set_attribute(node, "attr9", ["90", "90", "90"])
        self.set_attribute(
            node, "attr10", [onnx.numpy_helper.from_array(np.array([100]))]
        )
        self.set_attribute(
            node,
            "attr12",
            [
                onnx.helper.make_tensor_type_proto(
                    onnx.TensorProto.FLOAT, [120, 120, 120]
                ),
                onnx.helper.make_tensor_type_proto(
                    onnx.TensorProto.FLOAT, [120, 120, 120]
                ),
            ],
        )
        nonlocal passed
        passed = True

    rewriter = TestRewriter(
        SingleNodePattern("Dummy").with_domain("test"), repeat=1, test_func=_func
    )
    rewriter(_make_graph())
    assert passed


def test_match_twice():
    repeats = 0

    def _func(self, graph, nodes, *args, **kwargs):
        assert len(nodes) == 2
        nonlocal repeats
        repeats += 1

    pattern = GraphPattern()
    pattern.add_edge(
        SingleNodePattern("Dummy").with_domain("test"),
        SingleNodePattern("DummyTwo").with_domain("test"),
    )
    target_repeats = np.random.randint(2, 100)
    rewriter = TestRewriter(pattern, repeat=target_repeats, test_func=_func)
    rewriter(_make_graph(2))
    assert repeats == target_repeats


def test_match_invalid_pattern():
    class InvalidPattern(Pattern):
        def match(self, graph: OnnxGraph, specify_node_names=None):
            for node in graph.nodes:
                return graph.nodes[node]["pb"]

    pattern = InvalidPattern()
    rewriter = TestRewriter(pattern, 1, lambda *args, **kwargs: None)
    with pytest.raises(RuntimeError):
        rewriter(_make_graph())


def test_get_input_node():
    passed = False

    def _func(self, graph, nodes, *args, **kwargs):
        i1 = self.get_input_node(nodes[0], 0)
        i2 = self.get_input_node(nodes[0], -1)
        i3 = self.get_input_node(nodes[0], "n1")
        assert i1 is not None
        assert i2 is not None
        assert i1 is i2
        assert i2 is i3
        nonlocal passed
        passed = True

    rewriter = TestRewriter(SingleNodePattern("DummyTwo").with_domain("test"), 1, _func)
    rewriter(_make_graph(2))
    assert passed


def test_get_output_node():
    passed = False

    def _func(self, graph, nodes, *args, **kwargs):
        o1 = self.get_output_node(nodes[0], 0)[0]
        o2 = self.get_output_node(nodes[0], -1)[0]
        o3 = self.get_output_node(nodes[0], "n1")[0]
        assert o1 is not None
        assert o2 is not None
        assert o1 is o2
        assert o2 is o3
        nonlocal passed
        passed = True

    rewriter = TestRewriter(SingleNodePattern("Dummy").with_domain("test"), 1, _func)
    rewriter(_make_graph(3))
    assert passed


def test_get_value():
    cst = make_constant("cst", np.array([1], np.int64))
    add = make_node("Add", ["x", cst.output[0]], ["y"])
    shapeof = make_node("Shape", [cst.output[0]], ["z"])
    mul = make_node("Mul", ["y", "z"], ["w"])
    model = make_model(
        make_graph(
            [cst, add, shapeof, mul],
            "test",
            [],
            [make_tensor_value_info("w", onnx.TensorProto.INT64, [1])],
            [onnx.numpy_helper.from_array(np.array([2], np.int64), "x")],
        ),
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET],
    )
    graph = OnnxGraph(model)

    def _func(self, graph, nodes, *args, **kwargs):
        node = nodes[0]
        if node.op_type == "Add":
            cst = self.get_input_node(node, 1)
            assert (self.get_value(cst) == np.array([1], np.int64)).all()
            assert (self.get_value(node.input[1]) == np.array([1], np.int64)).all()
            assert (self.get_value("x") == np.array([2], np.int64)).all()
        else:
            assert node.op_type == "Mul"
            shapeof = self.get_input_node(node, 1)
            assert (self.get_value(shapeof) == np.array([1], np.int64)).all()
            result = self.get_value(node.input[0])
            assert np.all(result == np.array([3], np.int64))

    pattern = SingleNodePattern("Add") | SingleNodePattern("Mul")
    rewriter = TestRewriter(pattern, 1, _func)
    rewriter(graph)


def test_get_constant_values():
    cst_i = make_node("Constant", [], ["int"], name="const_int", value_int=1)
    cst_ints = make_node(
        "Constant", [], ["ints"], name="const_ints", value_ints=[1, 2, 3]
    )
    cst_f = make_node("Constant", [], ["float"], name="const_float", value_float=1.0)
    cst_floats = make_node(
        "Constant", [], ["floats"], name="const_floats", value_floats=[1.0, 2.0, 3.0]
    )
    cst_s = make_node("Constant", [], ["string"], name="const_string", value_string="1")
    cst_strings = make_node(
        "Constant", [], ["strings"], name="const_strings", value_strings=["1", "2", "3"]
    )
    model = make_model(
        make_graph(
            [cst_i, cst_ints, cst_f, cst_floats, cst_s, cst_strings],
            "test_cst",
            [],
            [
                make_tensor_value_info("int", onnx.TensorProto.INT64, []),
                make_tensor_value_info("ints", onnx.TensorProto.INT64, [3]),
                make_tensor_value_info("float", onnx.TensorProto.FLOAT, []),
                make_tensor_value_info("floats", onnx.TensorProto.FLOAT, [3]),
                make_tensor_value_info("string", onnx.TensorProto.STRING, []),
                make_tensor_value_info("strings", onnx.TensorProto.STRING, [3]),
            ],
        ),
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET],
    )
    graph = OnnxGraph(model)

    def _func(self, graph, nodes, *args, **kwargs):
        node = nodes[0]
        v = self.get_value_or_die(node.output[0])
        if node.name == "const_int":
            assert v.dtype == np.int64
            assert v == 1
        elif node.name == "const_ints":
            assert v.dtype == np.int64
            assert (v == np.array([1, 2, 3])).all()
        elif node.name == "const_float":
            assert v.dtype == np.float32
            assert v == 1.0
        elif node.name == "const_floats":
            assert v.dtype == np.float32
            assert (v == np.array([1.0, 2.0, 3.0])).all()
        elif node.name == "const_string":
            assert v.dtype == np.dtype("S1")
            assert v.item() == b"1"
        elif node.name == "const_strings":
            assert v.dtype == np.dtype("S1")
            assert (v == np.array([b"1", b"2", b"3"])).all()

    pattern = SingleNodePattern("Constant")
    rewriter = TestRewriter(pattern, 1, _func)
    rewriter(graph)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json

import numpy as np
import onnx
from onnx.helper import (
    make_function,
    make_graph,
    make_model,
    make_node,
    make_operatorsetid,
    make_tensor_value_info,
)

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.evaluator import Evaluator


def _make_graph():
    f_log = make_node("Log", ["a0"], ["a1"])
    f_sm = make_node("Softmax", ["a1"], ["a2"], axis=1)
    f_sm.attribute[0].ref_attr_name = "dim"
    fn = make_function(
        "test",
        "MyLogSoftmax",
        ["a0"],
        ["a2"],
        [f_log, f_sm],
        opset_imports=[ONNXIFIER_OPSET],
        attributes=["dim"],
    )
    # note: use a different attribute value
    log_softmax = make_node("MyLogSoftmax", ["X"], ["Y"], dim=2, domain="test")
    graph = make_graph(
        [log_softmax],
        "test",
        [make_tensor_value_info("X", onnx.TensorProto.FLOAT, [1, 2, 3])],
        [make_tensor_value_info("Y", onnx.TensorProto.FLOAT, [1, 2, 3])],
    )
    model = make_model(
        graph,
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET, make_operatorsetid("test", 1)],
        functions=[fn],
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def _make_graph_doc_string():
    f_log = make_node("Log", ["a0"], ["a1"])
    f_sm = make_node("Softmax", ["a1"], ["a2"], axis=1)
    f_sm.attribute[0].ref_attr_name = "dim"
    fn = make_function(
        "test",
        "MyLogSoftmax",
        ["a0"],
        ["a2"],
        [f_log, f_sm],
        opset_imports=[ONNXIFIER_OPSET],
        attributes=["dim"],
    )
    # note: use a different attribute value
    log_softmax = make_node(
        "MyLogSoftmax",
        ["X"],
        ["Y"],
        dim=2,
        domain="test",
        doc_string=json.dumps(["a", "b"]),
    )
    graph = make_graph(
        [log_softmax],
        "test",
        [make_tensor_value_info("X", onnx.TensorProto.FLOAT, [1, 2, 3])],
        [make_tensor_value_info("Y", onnx.TensorProto.FLOAT, [1, 2, 3])],
    )
    model = make_model(
        graph,
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET, make_operatorsetid("test", 1)],
        functions=[fn],
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def test_inline_functions():
    graph = _make_graph()
    runner1 = Evaluator(graph.model, "OnnxRuntime")

    pm = PassManager(
        ["inline_functions"], configs={"inline_functions": {"force": True}}
    )
    graph = pm.optimize(graph, strict=True)
    assert len(graph) == 2

    runner2 = Evaluator(graph.model, "OnnxRuntime")
    x = np.random.uniform(1, 10, size=[1, 2, 3]).astype(np.float32)
    y1 = runner1(["Y"], {"X": x})[0]
    y2 = runner2(["Y"], {"X": x})[0]
    assert np.allclose(y1, y2)


def test_inline_functions_doc_string():
    graph = _make_graph_doc_string()
    runner1 = Evaluator(graph.model, "OnnxRuntime")

    pm = PassManager(
        ["inline_functions"], configs={"inline_functions": {"force": True}}
    )
    graph = pm.optimize(graph, strict=True)
    assert set(graph) == {"a", "b"}

    runner2 = Evaluator(graph.model, "OnnxRuntime")
    x = np.random.uniform(1, 10, size=[1, 2, 3]).astype(np.float32)
    y1 = runner1(["Y"], {"X": x})[0]
    y2 = runner2(["Y"], {"X": x})[0]
    assert np.allclose(y1, y2)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=line-too-long
import numpy as np
from onnx import NodeProto, TensorProto
from onnx.helper import make_node, tensor_dtype_to_np_dtype

from ...utils import text_to_boolean
from .. import OnnxGraph, make_constant
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register("scaled_dot_product_attention")
class ScaledDotProductAttention(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/sequence/scaled-dot-product-attention.html

    https://pytorch.org/docs/stable/generated/torch.nn.functional.scaled_dot_product_attention.html

    Y = SDPA(X) = softmax(XW^T/sqrt(d_k))XW

    Ref:
        https://pytorch.org/tutorials/intermediate/scaled_dot_product_attention_tutorial.html
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto):
        inputs = self.get_input_nodes(ori_node)
        attention_mask, scale = None, None
        if len(inputs) == 4:
            attention_mask = inputs[3]
        elif len(inputs) == 5:
            attention_mask, scale = inputs[3:]
        causal = self.get_attribute(ori_node, "causal")
        if not isinstance(causal, str):
            raise ValueError(
                f"{ori_node.name} causal should be a string, but got {causal}"
            )
        is_causal = text_to_boolean(causal)

        dtype = tensor_dtype_to_np_dtype(graph.tensor_type(ori_node.input[0]))

        if scale is None:
            query_dim = graph.tensor_shape(ori_node.input[0])[-1]
            assert isinstance(query_dim, int) and query_dim > 0
            scale_value = (1 / np.sqrt(query_dim)).astype(dtype)
            scale = make_constant(f"{ori_node.name}/scale", scale_value)
            self += scale
        attn_bias = np.zeros((1, 1), dtype=dtype)
        if is_causal:
            L = graph.static_tensor_shape(ori_node.input[0])[-2]
            S = graph.static_tensor_shape(ori_node.input[1])[-2]
            attn_bias = np.triu(np.broadcast_to(-np.inf, (L, S)), k=1).astype(dtype)
        bias = make_constant(f"{ori_node.name}/bias", attn_bias)
        bias_result = bias.output[0]
        if not is_causal and attention_mask is not None:
            mask_dtype = graph.tensor_type(ori_node.input[3])
            if mask_dtype == TensorProto.BOOL:
                raise NotImplementedError("bool mask is not supported yet")
            else:
                # bias += attention_mask
                add_mask = make_node(
                    "Add",
                    inputs=[bias.output[0], attention_mask.output[0]],
                    outputs=[f"{ori_node.name}/add_attn_mask_output0"],
                    name=f"{ori_node.name}/add_attn_mask",
                )
                self += add_mask
                bias_result = add_mask.output[0]

        key_rank = len(graph.tensor_shape(ori_node.input[1]))
        key_perm = list(range(key_rank))
        key_perm[-2], key_perm[-1] = key_perm[-1], key_perm[-2]
        trans_node = make_node(
            "Transpose",
            inputs=[ori_node.input[1]],
            outputs=[f"{ori_node.name}/key/transpose_output0"],
            perm=key_perm,
            name=f"{ori_node.name}/key/transpose",
        )
        gemm0 = make_node(
            "MatMul",
            inputs=[ori_node.input[0], trans_node.output[0]],
            outputs=[f"{ori_node.name}/gemm0_output0"],
            name=f"{ori_node.name}/gemm0",
        )
        dot_scale = make_node(
            "Mul",
            inputs=[gemm0.output[0], scale.output[0]],
            outputs=[f"{ori_node.name}/qk_scale_output0"],
            name=f"{ori_node.name}/qk_scale",
        )
        add_bias = make_node(
            "Add",
            inputs=[dot_scale.output[0], bias_result],
            outputs=[f"{ori_node.name}/attn_bias_output0"],
            name=f"{ori_node.name}/attn_bias",
        )
        act_node = make_node(
            "Softmax",
            inputs=[add_bias.output[0]],
            outputs=[f"{ori_node.name}/softmax_output0"],
            axis=-1,
            name=f"{ori_node.name}/softmax",
        )
        gemm1 = make_node(
            "MatMul",
            inputs=[act_node.output[0], ori_node.input[2]],
            outputs=[ori_node.output[0]],
            name=f"{ori_node.name}",
        )

        self += [bias, trans_node, gemm0, dot_scale, add_bias, act_node]
        return gemm1



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

import random
import tempfile
from pathlib import Path

import numpy as np
import onnx
import onnx.external_data_helper
import onnx.helper as oh
import pooch
import pytest

from onnxifier import ONNXIFIER_OPSET, OnnxGraph, convert_graph
from onnxifier.utils import chdir


def _pick_random_nodes(graph: OnnxGraph, num_nodes: int = 1):
    indices = list(range(len(graph)))
    random.shuffle(indices)
    node_names = np.array(graph)[indices[:num_nodes]].tolist()
    if len(node_names) == 1:
        return node_names[0]
    return node_names


def test_loading_models(classification_models):
    model_url, hash_value = classification_models
    model = onnx.load_model(pooch.retrieve(model_url, hash_value))
    OnnxGraph(model)


def test_add_remove_node(classification_models):
    model_url, hash_value = classification_models
    model = onnx.load_model(pooch.retrieve(model_url, hash_value))

    graph = OnnxGraph(model)
    node_name = _pick_random_nodes(graph)
    node = graph.nodes[node_name]["pb"]
    graph.remove_onnx_node(node, no_replace=True)
    onnx.checker.check_model(graph.model)

    graph.add_onnx_node(node)
    onnx.checker.check_model(graph.model)

    graph.remove_onnx_node(node.name)
    try:
        onnx.checker.check_model(graph.model)
    except onnx.checker.ValidationError:
        pass
    else:
        raise RuntimeError("onnx.checker.ValidationError is expected to raise")

    graph.add_onnx_node(node)
    onnx.checker.check_model(graph.model)


def test_make_subgraph(classification_models):
    model_url, hash_value = classification_models
    model = onnx.load_model(pooch.retrieve(model_url, hash_value))

    graph = convert_graph(model, print_passes=False)
    model = onnx.version_converter.convert_version(graph.model, 19)
    graph = OnnxGraph(model)

    sub_node_names = _pick_random_nodes(graph, len(graph) // 5)
    sub_nodes = [graph.nodes[i]["pb"] for i in sub_node_names]

    sub_from_nodes = graph.onnx_subgraph(sub_nodes)
    sub_from_names = graph.onnx_subgraph(sub_node_names)
    onnx.checker.check_model(sub_from_nodes.model)
    onnx.checker.check_model(sub_from_names.model)


def test_tensor_info(classification_models):
    model_url, hash_value = classification_models
    model = onnx.load_model(pooch.retrieve(model_url, hash_value))

    graph = OnnxGraph(model)
    node_name = _pick_random_nodes(graph)
    node = graph.nodes[node_name]["pb"]
    for input_name in node.input:
        graph.tensor_info(input_name)
    for output_name in node.output:
        graph.tensor_info(output_name)


def test_model_save(classification_models):
    model_url, hash_value = classification_models
    model = onnx.load_model(pooch.retrieve(model_url, hash_value))

    graph = OnnxGraph(model)
    with tempfile.TemporaryDirectory() as tmpdir:
        graph.save(tmpdir + "/model")
        # text format is too slow
        # graph.save(tmpdir + "/model", format="textproto")
        # graph.save(tmpdir + "/model", format="json")
        # graph.save(tmpdir + "/model", format="onnxtxt")


def test_subgraph_outputs():
    def _make_test_model():
        """a  (b)
          |   /
          (Add)
          / |
        (x) y
        """
        node1 = onnx.helper.make_node(
            "Constant",
            [],
            ["a"],
            name="A",
            value=onnx.numpy_helper.from_array(np.zeros([16], "float32")),
        )
        add = onnx.helper.make_node("Add", ["a", "b"], ["c"], name="Add")
        id1 = onnx.helper.make_node("Identity", ["c"], ["x"], "ID1")
        id2 = onnx.helper.make_node("Identity", ["c"], ["y"], "ID2")
        graph = onnx.helper.make_graph(
            [node1, add, id1, id2],
            "test",
            [],
            [
                onnx.helper.make_value_info(
                    "x", onnx.helper.make_tensor_type_proto(1, [16])
                ),
                onnx.helper.make_value_info(
                    "y", onnx.helper.make_tensor_type_proto(1, [16])
                ),
            ],
            [onnx.numpy_helper.from_array(np.zeros([1], "float32"), "b")],
        )
        return onnx.helper.make_model(graph, opset_imports=[ONNXIFIER_OPSET])

    graph = OnnxGraph(_make_test_model())
    subgraph = graph.onnx_subgraph(["Add", "ID1"])
    assert len(subgraph) == 2
    assert len(subgraph.inputs) == 1
    assert len(subgraph.outputs) == 2


def test_subgraph_outputs_case2():
    def _make_test_model():
        """a   b
          |   /
          (Add)
          / |
        (x) c
        """
        add = onnx.helper.make_node("Add", ["a", "b"], ["c"], name="Add")
        id1 = onnx.helper.make_node("Identity", ["c"], ["x"], "ID1")
        graph = onnx.helper.make_graph(
            [add, id1],
            "test",
            [
                onnx.helper.make_value_info(
                    "a", onnx.helper.make_tensor_type_proto(1, [16])
                ),
                onnx.helper.make_value_info(
                    "b", onnx.helper.make_tensor_type_proto(1, [16])
                ),
            ],
            [
                onnx.helper.make_value_info(
                    "x", onnx.helper.make_tensor_type_proto(1, [16])
                ),
                onnx.helper.make_value_info(
                    "c", onnx.helper.make_tensor_type_proto(1, [16])
                ),
            ],
        )
        return onnx.helper.make_model(graph, opset_imports=[ONNXIFIER_OPSET])

    graph = OnnxGraph(_make_test_model())
    subgraph = graph.onnx_subgraph(["Add", "ID1"])
    assert len(subgraph) == 2
    assert len(subgraph.inputs) == 2
    assert len(subgraph.outputs) == 2


def test_onnx_node_hashable():
    node0 = onnx.helper.make_node("Add", ["x", "y"], ["z"], name="add0")
    node1 = onnx.helper.make_node("Add", ["x", "y"], ["z"], name="add1")
    node2 = onnx.helper.make_node("Sub", ["x", "y"], ["z"], name="add0")
    assert hash(node0) != hash(node1)
    # NOTE: hash value of onnx NodeProto is only determined by its name,
    #       type or attributes are not included.
    assert hash(node0) == hash(node2)


def test_add_duplicated_node():
    add0 = onnx.helper.make_node("Add", ["x", "y"], ["z"], name="add0")
    relu = onnx.helper.make_node("Relu", ["z"], ["w"], name="relu")
    model = onnx.helper.make_model(
        onnx.helper.make_graph(
            [add0, relu],
            "test_graph",
            [
                onnx.helper.make_tensor_value_info("x", onnx.TensorProto.FLOAT, [32]),
                onnx.helper.make_tensor_value_info("y", onnx.TensorProto.FLOAT, [32]),
            ],
            [onnx.helper.make_tensor_value_info("w", onnx.TensorProto.FLOAT, [32])],
        )
    )
    graph = OnnxGraph(model)
    # replace add0 with Sub
    new_add0 = onnx.helper.make_node("Sub", ["x", "y"], ["z"], name="add0")
    graph.add_onnx_node(new_add0)
    assert graph.nodes["add0"]["pb"].op_type == "Sub"


def test_add_node_override_outputs():
    add0 = onnx.helper.make_node("Add", ["x", "y"], ["z"], name="add0")
    act0 = onnx.helper.make_node("Relu", ["z"], ["w"], name="relu")
    act1 = onnx.helper.make_node("Sin", ["w"], ["u"], name="sin")
    model = onnx.helper.make_model(
        onnx.helper.make_graph(
            [add0, act0, act1],
            "test",
            [
                onnx.helper.make_tensor_value_info("x", onnx.TensorProto.FLOAT, [32]),
                onnx.helper.make_tensor_value_info("y", onnx.TensorProto.FLOAT, [32]),
            ],
            [
                onnx.helper.make_tensor_value_info("u", onnx.TensorProto.FLOAT, [32]),
            ],
        )
    )
    graph = OnnxGraph(model)
    new_add0 = onnx.helper.make_node("Add", ["x", "y"], ["w"], name="add0")
    graph.add_onnx_node(new_add0)
    graph.remove_onnx_node("relu")
    onnx.checker.check_model(graph.model, True)


def test_graph_property():
    g = onnx.helper.make_graph([], "test", [], [])
    m = onnx.helper.make_model(g, opset_imports=[ONNXIFIER_OPSET])
    graph = OnnxGraph(m)
    assert graph.name == "test"
    assert graph.ir_version == onnx.IR_VERSION
    assert graph.opset_version == ONNXIFIER_OPSET.version

    graph.opset_version = 10
    assert graph.opset_version == 10


def test_graph_save_as_external():
    g = onnx.helper.make_graph(
        nodes=[
            onnx.helper.make_node(
                "Constant",
                [],
                ["att_big"],
                value=onnx.numpy_helper.from_array(
                    np.ones([1024], dtype=np.float32), "att_big"
                ),
            ),
            onnx.helper.make_node(
                "Constant",
                [],
                ["att_small"],
                value=onnx.numpy_helper.from_array(
                    np.ones([1023], dtype=np.int8), "att_small"
                ),
            ),
        ],
        name="test",
        inputs=[],
        outputs=[],
        initializer=[
            onnx.numpy_helper.from_array(np.ones([1024], dtype=np.float32), "big"),
            onnx.numpy_helper.from_array(np.ones([1023], dtype=np.int8), "small"),
        ],
    )
    f = onnx.helper.make_function(
        "test",
        "foo",
        [],
        ["b"],
        [
            onnx.helper.make_node(
                "Constant",
                [],
                ["b"],
                value=onnx.numpy_helper.from_array(
                    np.ones([1024], dtype=np.float32), "b"
                ),
            )
        ],
        [ONNXIFIER_OPSET],
        ["attr_big"],
        [
            onnx.helper.make_attribute(
                "attr_big",
                onnx.numpy_helper.from_array(
                    np.ones([1024], dtype=np.float32), "attr_big"
                ),
            )
        ],
    )
    model = onnx.helper.make_model(g, opset_imports=[ONNXIFIER_OPSET], functions=[f])
    onnx.checker.check_model(model, True, True)
    graph = OnnxGraph(model)
    graph.restore_tensors_from_external()
    with tempfile.TemporaryDirectory() as tmpdir:
        # pylint: disable=protected-access
        graph.save(tmpdir + "/model.onnx", save_as_external_data=True)
        model = onnx.load_model(tmpdir + "/model.onnx", load_external_data=False)
        external_tensors = [
            t
            for t in onnx.external_data_helper._get_all_tensors(model)
            if onnx.external_data_helper.uses_external_data(t)
        ]
        # onnx enumerating tensors could be duplicated, because it yield function twice
        onnx.TensorProto.__hash__ = lambda x: hash(id(x))  # type: ignore
        # function attribute could be saved as external data
        assert len(set(external_tensors)) == 3

        with pytest.raises(FileNotFoundError):
            OnnxGraph(model, base_dir="wrong-base")
        with chdir(tmpdir):
            ext_graph = OnnxGraph(model)
        ext_data = list(ext_graph.external_data)
        assert len(ext_data) == 1
        assert ext_data[0] == Path(tmpdir) / "model"
        ext_graph.restore_tensors_from_external()
        external_tensors = [
            t
            for t in onnx.external_data_helper._get_all_tensors(ext_graph.model)
            if onnx.external_data_helper.uses_external_data(t)
        ]
        assert len(external_tensors) == 0
    # save to a new location because graph has been restored
    with tempfile.TemporaryDirectory() as tmpdir:
        ext_graph.save(tmpdir + "/model.onnx", save_as_external_data=True)


def test_graph_sibling_nodes():
    """Test get sibling nodes."""

    nodes = [
        oh.make_node("Identity", ["x"], ["y"], name="id1"),
        oh.make_node("Sin", ["y"], ["z"], name="sin1"),
        oh.make_node("Cos", ["y"], ["w"], name="cos1"),
    ]
    model = oh.make_model(
        oh.make_graph(
            nodes,
            "sibling_graph_test",
            inputs=[oh.make_tensor_value_info("x", onnx.TensorProto.FLOAT, [1])],
            outputs=[
                oh.make_tensor_value_info("z", onnx.TensorProto.FLOAT, [1]),
                oh.make_tensor_value_info("w", onnx.TensorProto.FLOAT, [1]),
            ],
        ),
        ir_version=onnx.IR_VERSION,
        opset_imports=[oh.make_operatorsetid("", 21)],
    )
    graph = OnnxGraph(model)
    sib1 = graph.onnx_siblings("sin1")
    sib2 = graph.onnx_siblings("cos1")
    assert len(sib1) == 1
    assert len(sib2) == 1
    assert sib1[0].name == "cos1"
    assert sib2[0].name == "sin1"



"""
Copyright (C) 2024-2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import inspect
import traceback
from collections import OrderedDict, defaultdict
from collections.abc import Mapping, Sequence, Set
from functools import partial
from itertools import chain
from typing import Any, Dict, List, Optional

import networkx as nx
from onnx import NodeProto, ValueInfoProto
from onnx.helper import make_function, make_value_info
from tabulate import tabulate
from termcolor import colored

from .graph import OnnxGraph
from .logger import debug, error, warning
from .passes import L1, L2, L3, PASSES
from .passes.utils import extract_function
from .traits import RewriterInterface


def _is_function(rewriter) -> bool:
    """Return True when ``rewriter`` is either a plain function or a
    ``functools.partial`` object that wraps a function.

    This simplifies repeated inline checks like:
    ``(isinstance(rewriter, partial) and
    inspect.isfunction(rewriter.func)) or inspect.isfunction(rewriter)``
    and centralises the logic for easier testing and future changes.
    """
    is_partial_func = False
    if isinstance(rewriter, partial) and inspect.isfunction(rewriter.func):
        is_partial_func = True
    is_plain_func = inspect.isfunction(rewriter)
    return is_partial_func or is_plain_func


def _apply_pass(rewriter, g, specify_node_names):
    if _is_function(rewriter) or not specify_node_names:
        g = rewriter(g)
    else:
        g = rewriter(g, _specify_node_names=specify_node_names)
    return g


class PassManager:
    """Ordered optimization pass list.

    Args:
        include (List[str | RewriterInterface], Optional): a list of pattern to
            select passes. Defaults to select L1, L2 and L3 passes.
        exclude (List[str], Optional): a list of pattern to deselect passes.
            Defaults to None.
        configs (Dict[str, Any], Optional): a dictionary of pass configurations.
            Defaults to None.
    """

    def __init__(
        self,
        include: Optional[Sequence[str | RewriterInterface]] = None,
        exclude: Optional[Sequence[str]] = None,
        configs: Optional[Dict[str, Any]] = None,
    ) -> None:
        passes: List[str | RewriterInterface]
        if include is None:
            passes = [i for i in chain(L1, L2, L3)]
        else:
            passes = [i for i in include]
        if exclude:
            passes = list(filter(lambda i: i not in exclude, passes))
        activated: List[RewriterInterface] = []
        for i in passes:
            if isinstance(i, str):
                if i in PASSES:
                    activated.append(PASSES[i])
                else:
                    warning(f"{i} is not registered as a pass, ignore it.")
            else:
                activated.append(i)
        self.activated: List[RewriterInterface] = activated
        if configs:
            self._assign_config_to_pass(configs)

    def _assign_config_to_pass(self, configs: Dict[str, Any]):
        for key, config in configs.items():
            index = -1
            if ":" in key:
                key, index_str = key.split(":", 2)
                index = int(index_str)
            if not isinstance(config, Mapping):
                warning(f"config {key}:{index} must be a dict, but got {type(config)}")
                continue
            candidates = [i for i in self.activated if i.__NAME__ == key]
            if index >= 0 and index >= len(candidates):
                warning(
                    f"config {key}:{index} exceeds the boundary. "
                    f"Number of {key} is {len(candidates)}"
                )
                continue
            if index >= 0:
                candidates = [candidates[index]]
            for func in candidates:
                pos = self.activated.index(func)
                self.activated[pos] = partial(func, **config)  # type: ignore
                self.activated[pos].__NAME__ = key
                self.activated[pos].__DEPS__ = func.__DEPS__
                self.activated[pos].__PATCHES__ = func.__PATCHES__

    def _expand(self, nodes, priv_member):
        root: nx.DiGraph = nx.DiGraph()  # type: ignore
        leaves = [f"{i}:{n}" for i, n in enumerate(nodes)]
        root.add_nodes_from(leaves)
        # a shallow graph to check cyclic dependencies
        shallow: nx.DiGraph = nx.DiGraph()  # type: ignore
        shallow.add_nodes_from(nodes)
        ind = len(leaves)
        while leaves:
            leaf = leaves.pop(0)
            leaf_pass = leaf.split(":")[1]
            children = getattr(PASSES.get(leaf_pass), priv_member, [])
            shallow.add_edges_from([(leaf_pass, child) for child in children])
            try:
                cycles = nx.find_cycle(shallow, leaf_pass)
            except nx.NetworkXNoCycle:
                children = [f"{ind + i}:{c}" for i, c in enumerate(children)]
                ind += len(children)
                root.add_edges_from([(leaf, child) for child in children])
                leaves.extend(children)
            else:
                error(f"Cyclic dependencies found!: {cycles}")
                raise RuntimeError("Cyclic dependencies found!")
        return root

    def _expand_deps(self, deps):
        root = self._expand(deps, "__DEPS__")
        for i in nx.traversal.dfs_postorder_nodes(root):
            yield i.split(":")[1]

    def _expand_patches(self, nodes):
        root = self._expand(nodes, "__PATCHES__")
        for i in nx.traversal.dfs_preorder_nodes(root):
            yield i.split(":")[1]

    def optimize(
        self,
        graph: OnnxGraph,
        strict: bool = False,
        recursive: bool = False,
        specify_node_names: Optional[Set[str]] = None,
    ) -> OnnxGraph:
        """Invoke passes on the input graph.

        Args:
            graph (OnnxGraph): See :class:`OnnxGraph`.
            strict (bool): Break if any pass fails.
            recursive (bool): Recursively apply passes to functions.
            specify_node_names (set[str], optional): Only optimize nodes with these
                names. Defaults to None.
        """
        graphs = {"": ("", graph)}  # let "" be the main graph
        if recursive:
            graphs.update(self._make_subgraph_from_functions(graph))
        for opt in self.activated:
            for name, (d, g) in graphs.items():
                try:
                    for deps in self._expand_deps(opt.__DEPS__):
                        if rewriter := PASSES.get(deps):
                            debug(f"Applying dependency pass: {deps} to {name}")
                            g = _apply_pass(rewriter, g, specify_node_names)
                    debug(f"Applying pass: {opt.__NAME__} to {name}")
                    g = _apply_pass(opt, g, specify_node_names)
                    for patch in self._expand_patches(opt.__PATCHES__):
                        debug(f"Applying patch pass: {patch} to {name}")
                        if rewriter := PASSES.get(patch):
                            g = _apply_pass(rewriter, g, specify_node_names)
                except Exception as ex:  # pylint: disable=broad-exception-caught
                    error(f"{opt.__NAME__} failed at {name}: {ex}")
                    debug("\n".join(traceback.format_exception(ex)))
                    if strict:
                        raise
                graphs[name] = (d, g)
                if name:
                    function = self._make_function_from_subgraph(name, d, g)
                    graphs[""][1].onnx_add_function(function, force=True)
        return graphs[""][1]

    def _make_subgraph_from_functions(self, graph: OnnxGraph):
        users: Dict[str, List[NodeProto]] = defaultdict(list)
        iter_order: Dict[str, OnnxGraph] = {}  # ordered hashset
        # graph.functions may not topologically sorted
        bfs: Dict[str, None] = OrderedDict()  # used as ordered set
        for n in graph:
            node: NodeProto = graph.nodes[n]["pb"]
            if node.op_type in graph.functions:
                users[node.op_type].append(node)
                iter_order[node.op_type] = graph
                bfs[node.op_type] = None
        while bfs:
            top_key = next(iter(bfs))
            f = graph.functions[top_key]
            del bfs[top_key]
            for node in f.node:
                if node.op_type in graph.functions:
                    users[node.op_type].append(node)
                    bfs[node.op_type] = None
                    iter_order[node.op_type] = extract_function(
                        graph,
                        f.node,
                        [(i, j) for i, j in zip(users[f.name][0].input, f.input)],
                        [(i, j) for i, j in zip(users[f.name][0].output, f.output)],
                    )
        extended_value_info: List[ValueInfoProto] = []
        for name, parent in iter_order.items():
            if len(users[name]) != 1:
                # Currently only function for a single node is OK to optimize
                # Otherwise we can't figure out how to do a general optimization
                # across different nodes.
                continue
            node = users[name][0]
            f = graph.functions[name]
            # pylint: disable=protected-access
            parent._value_info.extend(extended_value_info)
            model = extract_function(
                parent,
                f.node,
                [(i, j) for i, j in zip(node.input, f.input)],
                [(i, j) for i, j in zip(node.output, f.output)],
            )
            for i in chain(model.input, model.output):
                extended_value_info.append(make_value_info(i.name, i.type))
            yield name, (f.domain, model)

    def _make_function_from_subgraph(self, name: str, domain: str, model: OnnxGraph):
        # pylint: disable=protected-access
        model._keep_value_info = True
        h = model.model
        function = make_function(
            domain,
            name,
            list(model.inputs),
            list(model.outputs),
            h.graph.node,
            opset_imports=h.opset_import,
            value_info=h.graph.value_info,
        )
        return function

    @classmethod
    def print_all(cls):
        """Print the name of all passes."""
        print(PASSES, flush=True)

    @classmethod
    def print_l1(cls):
        """Print the name of all L1 passes."""
        print(L1, flush=True)

    @classmethod
    def print_l2(cls):
        """Print the name of all L2 passes."""
        print(L2, flush=True)

    @classmethod
    def print_l3(cls):
        """Print the name of all L3 passes."""
        print(L3, flush=True)

    @classmethod
    def print(cls, names: str | List[str]):
        """Print a specific pass or a set of passes."""
        print(PASSES.child(names), flush=True)

    def __repr__(self) -> str:
        return tabulate(
            [[i.__NAME__, i] for i in self.activated], ["PASS", "Func"], "grid"
        )


def print_pass_simple(pm: PassManager):
    """Print activated passes of a PassManager in a simple format."""

    msg = ""
    for n, i in enumerate(pm.activated):
        deps = []
        post = []
        # pylint: disable=protected-access
        for j in pm._expand_deps(i.__DEPS__):
            if j in PASSES:
                deps.append(j)
        for j in pm._expand_patches(i.__PATCHES__):
            if j in PASSES:
                post.append(j)
        msg += f"\n{n:<2} "
        if deps:
            msg += "[" + colored(",".join(deps), "yellow") + "] "
        msg += i.__NAME__
        if post:
            msg += " [" + colored(",".join(post), "magenta") + "]"
    if msg:
        print(f"Activated passes ([deps] pass [patches]):{msg}")



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

import json
from typing import List

from onnx.helper import make_function, make_node, make_operatorsetid
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register("fuse_swish")
class FuseSwishRewriter(Rewriter):
    r"""Fuse sigmoid-mul into Swish activation function.

     Swish is also known as SiLU (Sigmoid-Weighted Linear Unit) and is defined as:

    .. math::

         Swish(x) = x \cdot \sigma(x)

     where :math:`\sigma` is the sigmoid function.

     Before:

         a->sigmoid->mul->b
          \__________/

     After:

         a->swish(function)->b
    """

    def __init__(self):
        pattern = GraphPattern()
        node_a = SingleNodePattern()
        sigm = SingleNodePattern("Sigmoid")
        mul = SingleNodePattern("Mul")
        pattern.add_edge(node_a, sigm)
        pattern.add_edge(node_a, mul)
        pattern.add_edge(sigm, mul)
        super().__init__(pattern)

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        _, sigmoid, mul = nodes
        swish = make_node(
            "Swish",
            inputs=[sigmoid.input[0]],
            outputs=[mul.output[0]],
            name=f"{sigmoid.name}/Swish",
            domain="onnxifier",  # domain can not be empty or "ai.onnx"
            doc_string=json.dumps([sigmoid.name, mul.name]),
        )
        self += [swish]
        self -= [sigmoid, mul]

        swish_func = make_function(
            "onnxifier",
            "Swish",
            inputs=[sigmoid.input[0]],
            outputs=[mul.output[0]],
            nodes=[sigmoid, mul],
            opset_imports=[
                make_operatorsetid("", graph.opset_version),
                make_operatorsetid("onnxifier", 1),
            ],
        )
        graph.onnx_add_function(swish_func)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info
from onnx.numpy_helper import from_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.passes.fusion.group_op import _group_operators_recursively


def _make_graph():
    nodes = [
        make_node("Relu", inputs=["X"], outputs=["a0"], name="/m/act0/layer0/r0"),
        make_node("Relu", inputs=["a0"], outputs=["a1"], name="/m/act0/layer1/r1"),
        make_node("Add", inputs=["a1", "c2"], outputs=["a2"], name="/m/act0/layer2/r2"),
        make_node("Relu", inputs=["a2"], outputs=["a3"], name="/m/act1/layer3/r3"),
        make_node("Add", inputs=["a3", "c4"], outputs=["a4"], name="/m/act1/layer4/r4"),
        make_node("Relu", inputs=["a4"], outputs=["Y"], name="/m/act1/layer5/r5"),
    ]
    g = make_graph(
        nodes,
        name="hier_test",
        inputs=[
            make_tensor_value_info("X", onnx.TensorProto.FLOAT, [1024]),
        ],
        outputs=[
            make_tensor_value_info("Y", onnx.TensorProto.FLOAT, [1024]),
        ],
        initializer=[
            from_array(np.random.rand(1024).astype(np.float32), name="c2"),
            from_array(np.random.rand(1024).astype(np.float32), name="c4"),
        ],
    )
    model = make_model(
        g, opset_imports=[ONNXIFIER_OPSET], ir_version=ONNXIFIER_IR_VERSION
    )
    onnx.checker.check_model(model)
    return model


def test_group_operators_recursively():
    model = _make_graph()
    graph = OnnxGraph(model)
    hier = {
        "act0": {
            "layer0": "/m/act0/layer0/r0",
            "layer1": "/m/act0/layer1/r1",
            "layer2": "/m/act0/layer2/r2",
        },
        "act1": {
            "layer3": "/m/act1/layer3/r3",
            "layer4": "/m/act1/layer4/r4",
            "layer5": "/m/act1/layer5/r5",
        },
    }

    node = _group_operators_recursively(graph, hier["act0"], "act0")
    assert node.name == "act0"
    assert node.input[0] == "X"
    assert node.output[0] == "a2"
    assert "act0" in graph.functions
    assert graph.functions["act0"].name == "act0"
    assert "/m/act0/layer0/r0" not in graph
    assert "/m/act0/layer1/r1" not in graph
    assert "/m/act0/layer2/r2" not in graph
    assert "/m/act1/layer3/r3" in graph
    assert "/m/act1/layer4/r4" in graph
    assert "/m/act1/layer5/r5" in graph


def test_group_depth_1():
    model = _make_graph()
    graph = OnnxGraph(model)
    graph = PassManager(
        ["group"], configs={"group": {"depth": 1, "sep": "/"}}
    ).optimize(graph, strict=True)

    # depth=1: all nodes should be grouped under "m"
    assert "m" in graph.functions
    assert graph.functions["m"].name == "m"

    # all original nodes should be removed from the main graph
    assert "/m/act0/layer0/r0" not in graph
    assert "/m/act0/layer1/r1" not in graph
    assert "/m/act0/layer2/r2" not in graph
    assert "/m/act1/layer3/r3" not in graph
    assert "/m/act1/layer4/r4" not in graph
    assert "/m/act1/layer5/r5" not in graph

    # the function call node should exist
    assert "m" in graph
    m_node = graph.nodes["m"]["pb"]
    assert m_node.op_type == "m"
    assert m_node.input[0] == "X"
    assert m_node.output[0] == "Y"


def test_group_depth_2():
    model = _make_graph()
    graph = OnnxGraph(model)
    graph = PassManager(
        ["group"], configs={"group": {"depth": 2, "sep": "/"}}
    ).optimize(graph, strict=True)

    # depth=2: nodes should be grouped under "m" -> "act0"/"act1"
    assert "m" in graph.functions
    assert "act0" in graph.functions
    assert "act1" in graph.functions

    # all original nodes should be removed from the main graph
    assert "/m/act0/layer0/r0" not in graph
    assert "/m/act0/layer1/r1" not in graph
    assert "/m/act0/layer2/r2" not in graph
    assert "/m/act1/layer3/r3" not in graph
    assert "/m/act1/layer4/r4" not in graph
    assert "/m/act1/layer5/r5" not in graph

    # the top-level function call node should exist
    assert "m" in graph
    m_node = graph.nodes["m"]["pb"]
    assert m_node.op_type == "m"
    assert m_node.input[0] == "X"
    assert m_node.output[0] == "Y"


def test_group_depth_3():
    # depth=3: nodes should be grouped under
    # "m" -> "act0"/"act1" -> "layer0"/"layer1"/...
    model = _make_graph()
    graph = OnnxGraph(model)
    graph = PassManager(
        ["group"], configs={"group": {"depth": 3, "sep": "/"}}
    ).optimize(graph, strict=True)

    assert "m" in graph.functions
    assert "act0" in graph.functions
    assert "act1" in graph.functions
    assert "layer0" in graph.functions
    assert "layer1" in graph.functions
    assert "layer2" in graph.functions
    assert "layer3" in graph.functions
    assert "layer4" in graph.functions
    assert "layer5" in graph.functions

    # all original nodes should be removed from the main graph
    assert "/m/act0/layer0/r0" not in graph
    assert "/m/act0/layer1/r1" not in graph
    assert "/m/act0/layer2/r2" not in graph
    assert "/m/act1/layer3/r3" not in graph
    assert "/m/act1/layer4/r4" not in graph
    assert "/m/act1/layer5/r5" not in graph

    # the top-level function call node should exist
    assert "m" in graph
    m_node = graph.nodes["m"]["pb"]
    assert m_node.op_type == "m"
    assert m_node.input[0] == "X"
    assert m_node.output[0] == "Y"



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import tempfile
from pathlib import Path

import onnx
from openvino import serialize

from onnxifier.domain.intel.openvino.xml_frontend import openvino_xml_to_onnx_graph


def convert_xml(model, opset_version=None):
    """test model conversion from openvino xml to onnx"""
    with tempfile.TemporaryDirectory() as tmpdir:
        serialize(model, Path(tmpdir) / "model.xml", Path(tmpdir) / "model.bin")
        model = openvino_xml_to_onnx_graph(Path(tmpdir) / "model.xml")
        if opset_version is not None:
            model.opset_import[0].domain = ""
            model.opset_import[0].version = opset_version
        onnx.checker.check_model(model, True)
        onnx.save_model(model, Path(tmpdir) / "model.onnx")
    return model



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_operatorsetid,
    make_tensor_value_info,
)

from onnxifier import OnnxGraph, PassManager


def _build_graph(opset=19, scales=[1, 1, 1, 1, 1]):
    resize = make_node(
        "Resize", ["x", "", "scales"], ["y"], axes=[2, 3, 4], name="resize"
    )
    graph = make_graph(
        [resize],
        "test",
        [make_tensor_value_info("x", onnx.TensorProto.FLOAT, [1, 3, 64, 128, 256])],
        [
            make_tensor_value_info(
                "y", onnx.TensorProto.FLOAT, ["n", "c", "d", "h", "w"]
            )
        ],
        [onnx.numpy_helper.from_array(np.array(scales, np.float32), "scales")],
    )
    model = make_model(graph, opset_imports=[make_operatorsetid("", opset)])
    return OnnxGraph(model)


def test_not_supported_opset():
    graph = _build_graph(opset=17)
    # do nothing if opset is less than 18
    assert graph.opset_version == 17
    pm = PassManager(["remove_resize_batch_and_channel_scales"])
    pm.optimize(graph, strict=True)


def test_remove_resize_batch_and_channel_scales():
    graph = _build_graph(opset=19, scales=(1, 1, 0.5, 0.5, 0.5))
    pm = PassManager(["remove_resize_batch_and_channel_scales", "infer_shape"])
    graph = pm.optimize(graph, strict=True)
    assert graph.tensor_shape("scales/new_output_0") == [3]
    onnx.checker.check_model(graph.model)
    onnx.shape_inference.infer_shapes(graph.model, True, True)

    graph = _build_graph(opset=19, scales=(1, 2, 2, 2, 1))
    pm = PassManager(["remove_resize_batch_and_channel_scales", "infer_shape"])
    graph = pm.optimize(graph, strict=True)
    assert graph.tensor_shape("scales/new_output_0") == [4]
    onnx.checker.check_model(graph.model)
    onnx.shape_inference.infer_shapes(graph.model, True, True)

    graph = _build_graph(opset=19, scales=(2, 2, 1, 1, 1))
    pm = PassManager(["remove_resize_batch_and_channel_scales", "infer_shape"])
    graph = pm.optimize(graph, strict=True)
    assert graph.tensor_shape("scales/new_output_0") == [5]
    onnx.checker.check_model(graph.model)
    onnx.shape_inference.infer_shapes(graph.model, True, True)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import io

import numpy as np
import onnx
import pytest

from onnxifier.evaluator import Evaluator

torch = pytest.importorskip("torch")
nn = pytest.importorskip("torch.nn")


class Diag(nn.Module):
    def forward(self, x):
        return torch.diag(x, diagonal=0)


def test_diag_export_without_onnxifier():
    """Pytorch has no implementation of torch.diag for onnx."""

    diag = Diag()
    f = io.BytesIO()
    with pytest.raises(RuntimeError):
        torch.onnx.export(diag, (torch.randn(3, 3),), f, opset_version=17, dynamo=False)


def test_diag_export_in_onnxifier():
    """Register aten::diag symbolic by import onnxifier"""
    # pylint: disable=import-outside-toplevel, unused-import
    import onnxifier.domain.pytorch.diag  # noqa: F401

    diag = Diag()
    f = io.BytesIO()
    x = torch.randn(3)
    torch.onnx.export(diag, (x,), f, opset_version=17, dynamo=False)

    f.seek(0)
    model = onnx.load(f)
    assert model.graph.node[0].op_type == "diag"

    runner = Evaluator(model, "onnxruntime")

    y = diag(x).detach().numpy()
    z = runner([], {model.graph.input[0].name: x.numpy()})[0]

    assert np.allclose(y, z)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_value_info,
)

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.passes.utils import evaluate_on_node


def _build_test_graph():
    # Create a constant tensor as input
    input_tensor = make_tensor(
        "input_value",
        onnx.TensorProto.FLOAT,
        [2, 3, 4],
        np.random.rand(2, 3, 4).astype(np.float32),
    )

    # Create a Shape node that takes the input
    shape_node = make_node("Shape", ["input"], ["shape_output"], name="shape_node")

    # Build the graph
    graph = make_graph(
        [shape_node],
        "shape_test_graph",
        [make_tensor_value_info("input", onnx.TensorProto.FLOAT, [2, 3, 4])],
        [make_tensor_value_info("shape_output", onnx.TensorProto.INT64, [3])],
        initializer=[input_tensor],
    )

    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return model


def _build_test_graph_with_attrs():
    # Create a constant tensor as input
    input_tensor = make_tensor(
        "input_value",
        onnx.TensorProto.FLOAT,
        [2, 3, 4, 5],
        np.random.rand(2, 3, 4, 5).astype(np.float32),
    )

    # Create a Shape node with start and end attributes
    shape_node = make_node(
        "Shape", ["input"], ["shape_output"], name="shape_node", start=1, end=3
    )

    # Build the graph
    graph = make_graph(
        [shape_node],
        "shape_attr_test_graph",
        [make_tensor_value_info("input", onnx.TensorProto.FLOAT, [2, 3, 4, 5])],
        [make_tensor_value_info("shape_output", onnx.TensorProto.INT64, [2])],
        initializer=[input_tensor],
    )

    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return model


def test_shape_to_constant_basic():
    model = _build_test_graph()
    graph = OnnxGraph(model)
    pm = PassManager(["shape_to_constant"])
    graph = pm.optimize(graph, strict=True)

    # After optimization, there should be no Shape node
    assert len(graph.nodes) == 1
    for node_name in graph.nodes:
        node = graph.nodes[node_name]["pb"]
        # The Shape node should be replaced with Constant
        assert node.op_type == "Constant"
        # Check if the constant values are correct ([2, 3, 4])
        const_value = evaluate_on_node(graph, node)
        assert const_value is not None
        np.testing.assert_array_equal(const_value, np.array([2, 3, 4], dtype=np.int64))


def test_shape_to_constant_with_attributes():
    model = _build_test_graph_with_attrs()
    graph = OnnxGraph(model)
    pm = PassManager(["shape_to_constant"])
    graph = pm.optimize(graph, strict=True)

    # After optimization, there should be no Shape node
    assert len(graph.nodes) == 1
    for node_name in graph.nodes:
        node = graph.nodes[node_name]["pb"]
        # The Shape node should be replaced with Constant
        assert node.op_type == "Constant"
        # Check if the constant values are correct ([3, 4]) - sliced with start=1, end=3
        const_value = evaluate_on_node(graph, node)
        assert const_value is not None
        np.testing.assert_array_equal(const_value, np.array([3, 4], dtype=np.int64))
